# TODO: this module is totally incomplete
import re
pattern_type = type(re.compile(''))


class _Patterns:
    name = re.compile(r'\w+')


def is_type_pattern(o):
    return type(o) == pattern_type


class SyLexer:
    PATTERN_TUPLES = ()

    def __init__(self):
        assert isinstance(self.PATTERN_TUPLES, (type(None), list, tuple))
        patterns = [] if not isinstance(self.PATTERN_TUPLES, (tuple, list)) else list(self.PATTERN_TUPLES)

        self._patterns = []
        self.__pattern_names = set()

        for pattern in patterns:
            assert isinstance(pattern, (list, tuple)) and len(pattern) == 2
            self.add_pattern(pattern[0], pattern[1])

    def add_pattern(self, token_name, pattern):
        assert token_name not in self.__pattern_names
        assert _Patterns.name.match(pattern[0])
        token_tuple = (token_name, pattern)
        self._patterns.append(token_tuple)
        self.__pattern_names.add(token_name)

    def lex(self, text):
        assert isinstance(text, str)
        if len(text) == 0:
            return ()
        pos = 0


class SyToken:
    def __init__(self, name, value, match, start, end):
        self.name = name
        self.value = value
        self.match = match
        self.start = start
        self.end = end


class SyRule:
    def __init__(self, name, patterns):
        assert isinstance(name, str)
        self._name = name
        self._patterns = []
        self.__pattern_names = set()
        if not isinstance(patterns, (list, tuple)):
            assert isinstance(patterns, (str, pattern_type))

    @property
    def name(self):
        return self._name

    def add_pattern(self, token_name, pattern):
        assert token_name not in self.__pattern_names
        assert _Patterns.name.match(pattern[0])
        token_tuple = (token_name, pattern)
        self._patterns.append(token_tuple)
        self.__pattern_names.add(token_name)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
from synamic.core.contracts.content import ContentContract, CDocType


class StaticContent(ContentContract):
    def __init__(self, site, cfields):
        self.__site = site
        self.__cfields = cfields

        # validation
        assert CDocType.is_binary(self.cfields.cdoctype)
        assert cfields.cpath is not None

    @property
    def site(self):
        return self.__site

    def get_stream(self):
        file = self.__cfields.cpath.open('rb')
        return file

    @property
    def body(self):
        # TODO: body should not be allowed for binary files. Fix it later.
        with self.get_stream() as f:
            res = f.read()
        return res

    @property
    def body_as_string(self):
        return self.body.decode('utf-8')

    @property
    def body_as_bytes(self):
        return self.body

    @property
    def cfields(self):
        return self.__cfields

    def __getitem__(self, key):
        return self.__cfields[key]

    def __getattr__(self, key):
        return getattr(self.__cfields, key)



import mimetypes
import re
from PIL import Image
from synamic.core.contracts.content import ContentContract, CDocType
from synamic.core.synamic.router.url import ContentUrl
from io import BytesIO


def construct_new_file_name_comps(path, width, height):
    basename = path.basename
    dircomps = path.dirname_comps

    m = re.search(r'(\.jpg|\.png|\.gif|\.jpeg)$', basename, re.I)
    if m:
        base = basename[:m.start()]
        ext = basename[m.start():m.end()]
        new_name = base + str(width) + 'x' + str(height) + ext
    else:
        new_name = basename + str(width) + 'x' + str(height)
    res = (*dircomps, new_name)
    return res


class ResizedImageContent(ContentContract):
    def __init__(self, site, original_image_path, width, height):
        self.__site = site
        self.__original_image_path = original_image_path
        self.__width = int(width)
        self.__height = int(height)
        self.__url = None

        self.__file_name_comps = construct_new_file_name_comps(self.__original_image_path, self.__width, self.__height)
        self.__new_file_path = self.__site.path_tree.create_cpath(self.__file_name_comps, is_file=True)

    @property
    def original_image_path(self):
        return self.__original_image_path

    @property
    def width(self):
        return self.__width

    @property
    def height(self):
        return self.__height

    @property
    def new_file_path(self):
        return self.__new_file_path

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.__original_image_path == other.original_image_path and self.__width == other.width and self.__height == other.height

    def __hash__(self):
        return hash(self.__file_name_comps)

    @property
    def site(self):
        return self.__site

    @property
    def path(self):
        return self.new_file_path

    @property
    def cpath(self):
        return self.path

    @property
    def id(self):
        return None

    def get_stream(self):
        img = Image.open(self.__original_image_path.abs_path)
        new_img = img.resize((self.width, self.height), Image.BICUBIC)
        bio = BytesIO()
        format = img.format
        new_img.save(bio, format)
        new_img.close()
        img.close()
        del new_img
        del img
        bio.seek(0)
        return bio

    @property
    def mimetype(self):
        mimetype = 'octet/stream'
        path = self.__url.to_file_system_path
        type, enc = mimetypes.guess_type(path)
        if type:
            mimetype = type
        return mimetype

    @property
    def cdoctype(self):
        return ContentContract.__document_types.AUXILIARY

    @property
    def curl(self):
        if self.__url is None:
            self.__url = ContentUrl(self.__site, self.__new_file_path)
        return self.__url



from collections import OrderedDict
parent_indicator = 'parent:'

_parent_indicators = OrderedDict((
    ('p:', 'p'),
    ('parent:', 'parent'),
    ('r:', 'r'),
    ('root:', 'root'),
))


def parent_config_str_splitter(text: str):
    if text.startswith(parent_indicator):
        from_parent = True
        res = (from_parent, text[len(parent_indicator):])
    else:
        from_parent = False
        res = (from_parent, text)
    return res



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
from synamic.core.standalones.functions.decorators import not_loaded
from synamic.core.contracts import CDocType


class UserService:
    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False
        self.__menu_dir = self.__site.system_settings['dirs.metas.users']

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        # for file_path in file_paths:
        self.__is_loaded = True

    def make_user(self, user_id):
        path_tree = self.__site.get_service('path_tree')
        fn = user_id + '.syd'
        path = path_tree.create_file_cpath(self.__menu_dir, fn)
        syd = self.__site.object_manager.get_syd(path)

        user_obj = self.__User(
            self.__site,
            user_id,
            syd
        )
        return user_obj

    def get_user_ids(self):
        user_ids = []
        path_tree = self.__site.get_service('path_tree')
        users_cdir = path_tree.create_dir_cpath(self.__menu_dir)
        if users_cdir.exists():
            user_cfiles = users_cdir.list_files(depth=1)
            for menu_cfile in user_cfiles:
                basename = menu_cfile.basename
                if basename.lower().endswith('.syd'):
                    menu_name = basename[:-len('.syd')]
                    user_ids.append(menu_name)
        return tuple(user_ids)

    @classmethod
    def is_type_user(cls, other):
        return isinstance(other, cls.__User)

    class __User:
        def __init__(self, site, user_id, user_fields):
            self.__site = site
            self.__user_id = user_id.lower()
            self.__user_fields = user_fields

            # validation
            assert isinstance(self.__user_id, str)

            # calculate name
            name = self.__user_fields.get('name', None)
            if name is None:
                self.__user_fields.set('name', self.__user_id)
            self.__name = name

            # content
            self.__synthetic_cfields = None
            self.__content = None

        @property
        def id(self):
            return self.__user_id

        @property
        def name(self):
            return self.__name

        @property
        def cfields(self):
            if self.__synthetic_cfields is not None:
                sf = self.__synthetic_cfields
            else:
                system_settings = self.__site.system_settings
                content_service = self.__site.get_service('contents')
                url_partition_comp = system_settings['url_partition_comp']
                user_url_comp = system_settings['user_url_comp']
                cdoctype = CDocType.GENERATED_HTML_DOCUMENT
                mimetype = 'text/html'
                curl = self.__site.synamic.router.make_url(
                    self.__site, f'/{url_partition_comp}/{user_url_comp}/{self.id}', for_cdoctype=cdoctype
                )
                sf = content_service.make_synthetic_cfields(
                    curl,
                    cdoctype,
                    mimetype,
                    cpath=None,
                    fields_map=None)
                sf['title'] = self.title if self.title else self.name
                sf['author'] = self
                self.__synthetic_cfields = sf
            return sf

        @property
        def curl(self):
            return self.cfields.curl

        def __user_content_renderer(self, site, gen_content):
            site_settings = site.settings
            template_service = site.get_service('templates')
            user_template_name = site_settings['templates.user']

            html_text_content = template_service.render(
                user_template_name,
                site=site,
                content=gen_content,
                author=self,
            )
            return html_text_content

        @property
        def content(self):
            if self.__content is not None:
                content = self.__content
            else:
                content_service = self.__site.get_service('contents')

                content = content_service.build_generated_content(
                    self.cfields,
                    '',
                    source_cpath=None,
                    render_callable=self.__user_content_renderer)
                self.__content = content
            return content

        @property
        def contents(self):
            return self.__site.object_manager.query_contents(f'author == {self.id} :sortby created_on desc')

        def __getattr__(self, item):
            return self.__user_fields.get(item, None)

        def __getitem__(self, item):
            return self.__user_fields.get(item, None)

        def __eq__(self, other):
            if not isinstance(other, self.__class__):
                return False
            return self.id == other.id

        def __hash__(self):
            return hash(self.id)

        def __str__(self):
            return "User (%s) : %s" % (self.id, self.name)

        def __repr__(self):
            return repr(str(self))




"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


import os
from synamic.core.synamic import Synamic
from synamic.entry_points.dev_server.server import serve

site_root = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_site")

print("SITE-ROOT: %s" % site_root)

syn = Synamic(site_root)
syn.load()
syn.initialize_site()
serve(syn, 8000)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import os
import collections
import jinja2.exceptions

__all__ = [
    # functions
    'get_source_snippet_from_file', 'get_source_snippet_from_text',
    # exception classes
    'SynamicError', 'SynamicErrors', 'SynamicTemplateError', 'SynamicQueryParsingError', 'SynamicGetCParsingError',
    'SynamicGetCError', 'SynamicPreProcessorNotFound', 'SynamicMarkerNotFound', 'SynamicMarkNotFound',
    'SynamicInvalidNumberFormat', 'SynamicModelParsingError', 'SynamicInvalidDateTimeFormat',
    'SynamicSettingsError', 'SynamicInvalidCPathComponentError', 'SynamicPathDoesNotExistError',
    'SynamicSydParseError', 'SynamicFSError', 'SynamicDataError', 'SynamicMarkerIsNotPublic', 'SynamicSiteNotFound',
    'SynamicUserNotFound',
]


def format_msg_map(msg_map):
    lines = []
    for key, value in msg_map.items():
        if isinstance(value, str) and '\n' in value:
            value = '\n' + value
        lines.append(
            f'{key}: {value}'
        )
    if lines:
        lines.append('\n')
    return '\n'.join(lines)


def get_source_snippet_from_file(fn, line_no, limit=10):
    if os.path.exists(fn):
        with open(fn, encoding='utf-8') as f:
            source = f.read()
    else:
        source = ''
    return get_source_snippet_from_text(source, line_no, limit=limit)


def get_source_snippet_from_text(text, line_no, limit=10):
    """Line no starts at 1 not at 0"""
    source = text
    lines = source.splitlines()

    half_limit = limit//2
    half_limit_l = half_limit
    if line_no > half_limit:
        half_limit_l = line_no - half_limit
    else:
        half_limit_l = 1
    res = []
    res.extend(
        ['.... ' + line for line in lines[half_limit_l - 1:line_no - 1]]
    )

    error_line = lines[line_no - 1]
    res.append('->.. ' + error_line)

    res.extend(
        ['.... ' + line for line in lines[line_no:line_no + half_limit]]
    )
    return '\n'.join(res)


class SynamicError(Exception):
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return self.message

    def __repr__(self):
        return repr(self.__str__())


class SynamicErrors(SynamicError):
    def __init__(self, message, *synamic_errors):
        messages_for_errors = ''
        for err in synamic_errors:
            assert isinstance(err, SynamicError)
            messages_for_errors += f'\n{err.__class__.__name__}:\n{"-" * len(err.__class__.__name__)}\n' + err.message

        self.message = message + '\n' + messages_for_errors
        self.errors = synamic_errors


class SynamicTemplateError(SynamicError):
    def __init__(self, jinja_ex):
        assert isinstance(jinja_ex, jinja2.exceptions.TemplateError),\
            f'Exception instance passed to {self.__class__.__name__} must be of jinja2.TemplateError'
        self.error_map = error_map = collections.OrderedDict()
        self.error_type = None
        self.message = None

        if isinstance(jinja_ex, jinja2.exceptions.TemplateSyntaxError):
            self.error_type = 'Template Syntax Error'
            self.message = self.message
            error_map['Error Message'] = jinja_ex.message
            error_map['Line No'] = jinja_ex.lineno
            error_map['Template Name'] = jinja_ex.name
            error_map['File Name'] = jinja_ex.filename
            error_map['Source'] = get_source_snippet_from_file(jinja_ex.filename, jinja_ex.lineno, limit=10)
        elif isinstance(jinja_ex, jinja2.exceptions.TemplateRuntimeError):
            self.error_type = 'Template Runtime Error'
            self.message = self.message
        elif isinstance(jinja_ex, jinja2.exceptions.TemplatesNotFound):
            self.error_type = 'Templates not Found'
            self.message = self.message
            error_map['Template Names'] = jinja_ex.names
        elif isinstance(jinja_ex, jinja2.exceptions.TemplateNotFound):
            self.error_type = 'A Template not Found'
            self.message = self.message
            error_map['Template Name'] = jinja_ex.name
        elif isinstance(jinja_ex, jinja2.exceptions.UndefinedError):
            self.error_type = 'Template Undefined Variable'
            self.message = self.message
        else:
            raise Exception('What have i missed!?')

        self.message = \
\
f"""
Error Type: {self.error_type}
{format_msg_map(self.error_map)}
Details:
{str(jinja_ex)}
"""


class SynamicQueryParsingError(SynamicError):
    """Raised when there is an error in lexing or parsing query string."""


class SynamicGetCParsingError(SynamicError):
    """Raised when there is an error in lexing or parsing param string of getc()"""


class SynamicGetCError(SynamicError):
    """Desired result was not found with SynamicGetCError"""


class SynamicPreProcessorNotFound(SynamicError):
    """Raised when pre processor not found"""


class SynamicMarkerNotFound(SynamicError):
    """Marker not found"""


class SynamicMarkNotFound(SynamicError):
    """Mark Not found"""


class SynamicMarkerIsNotPublic(SynamicError):
    """When marker is not public"""


class SynamicInvalidNumberFormat(SynamicError):
    """When number format is invalid"""


class SynamicInvalidDateTimeFormat(SynamicError):
    """When date, time or datetime is invalid"""


class SynamicModelParsingError(SynamicError):
    """When model cannot be parsed properly"""


class SynamicSettingsError(SynamicError):
    """Error related to settings and settings values"""


class SynamicInvalidCPathComponentError(SynamicError):
    """When a component passed to the cpath processing code is invalid."""


class SynamicPathDoesNotExistError(SynamicError):
    """When the path is non existent"""


class SynamicSydParseError(SynamicError):
    """Syd parse error in curlybrace parser."""


class SynamicFSError(SynamicError):
    """Synamic file system error"""


class SynamicDataError(SynamicError):
    """Exception for data in data service"""


class SynamicSiteNotFound(SynamicError):
    """Site not found"""


class SynamicUserNotFound(SynamicError):
    """User not found"""


class LogicalError(SynamicError):
    pass


class InvalidQueryString(SynamicError):
    pass


class GetUrlFailed(SynamicError):
    pass


class GetContentFailed(SynamicError):
    pass


class ParsingError(SynamicError):
    pass


class InvalidFrontMatter(SynamicError):
    pass


class InvalidFileNameFormat(SynamicError):
    pass


class DuplicateContentId(SynamicError):
    pass



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
from synamic.core.standalones.functions.decorators import not_loaded, loaded
from synamic.core.contracts import CDocType
from synamic.core.synamic.router import RouterService


class SitemapService:
    def __init__(self, site):
        self.__site = site
        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    @loaded
    def make_sitemap(self, all_cfields, url_str='sitemap.xml', mimetype='text/xml', cdoctype=CDocType.GENERATED_TEXT_DOCUMENT):
        content_service = self.__site.get_service('contents')
        url_texts = []
        for cfields in all_cfields:
            url_texts.append(self.__url_template(cfields))
        doc_text = self.__doc_template('\n'.join(url_texts))

        curl = RouterService.make_url(self.__site, url_str, for_cdoctype=cdoctype)
        synthetic_cfields = content_service.make_synthetic_cfields(curl, cdoctype, mimetype)
        gen_content = content_service.build_generated_content(synthetic_cfields, doc_text)
        return gen_content

    def __doc_template(self, urls_text):
        return \
f"""<?xml version="1.0" encoding="UTF-8"?>
<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9">
{urls_text}
</urlset>  
"""

    def __url_template(self, cfields):
        return \
f"""
    <url>
        <loc>{ cfields.curl.url }</loc>
        <lastmod>{ cfields.updated_on.strftime('%Y-%m-%d') }</lastmod>
        <changefreq>daily</changefreq>
        <priority>0.9</priority>
    </url>
    
"""



import re
from jinja2 import BaseLoader, TemplateNotFound
from synamic.exceptions import SynamicSiteNotFound

template_name_pat = re.compile(
    r"""^(?P<site_id>(
                        [^\s:/\\]+(~[^\s:/\\]+)+|~[^\s:/\\]+|~
                     )/
          )?
        (?P<path>.+)"""
    , re.VERBOSE)
# sync ~ in pattern system_settings['configs.site_id_sep']


class SynamicJinjaFileSystemLoader(BaseLoader):
    def __init__(self, site):
        self.__site = site

    def get_source(self, environment, template):
        site_id_sep = self.__site.synamic.system_settings['configs.site_id_sep']
        template_dir = self.__site.system_settings['dirs.templates.templates']
        site = self.__site
        template_name_match = template_name_pat.match(template)
        site_id = template_name_match.group('site_id')
        template_name = template_name_match.group('path')

        # only load from the current site.
        if site_id == site_id_sep:
            site = self.__site
        elif site_id is not None:
            site_id_obj = self.__site.synamic.sites.make_id(site_id)
            try:
                site = self.__site.synamic.sites.get_by_id(site_id_obj)
            except KeyError:
                raise SynamicSiteNotFound(
                    f'Site with id {site_id} was not found when looking for templated named {template} ->'
                    f' {template_name} inside of it.'
                )
        else:
            path_tree = site.path_tree
            template_cdir = path_tree.create_dir_cpath(template_dir)
            template_cfile = template_cdir.join(template, is_file=True)

            while True:
                path_tree = site.path_tree
                template_cdir = path_tree.create_dir_cpath(template_dir)
                template_cfile = template_cdir.join(template, is_file=True)
                if template_cfile.exists():
                    break
                else:
                    if site.has_parent:
                        site = site.parent
                        continue
                    else:
                        break
        path_tree = site.path_tree
        template_cdir = path_tree.create_dir_cpath(template_dir)
        template_cfile = template_cdir.join(template_name, is_file=True)

        if not template_cfile.exists():
            raise TemplateNotFound(
                template,
                message=f"Template "
                        f"{template_cfile.abs_path} does not exits for site with id {site.id.as_string}"
            )

        with template_cfile.open('r', encoding='utf-8') as f:
            source = f.read()
        last_gmtime = template_cfile.getmtime()
        return source, template_cfile.abs_path, lambda: template_cfile.getmtime() == last_gmtime



import re
from synamic.core.contracts import CDocType
from synamic.exceptions import SynamicMarkerIsNotPublic

_mark_title2id_sub_pat = re.compile(r'[\s:-]')

separator_comma_pat = re.compile(r',[^,]?')


class _Mark:
    def __init__(self, parent, site, mark_map, marker):
        if mark_map.get('title', None) is None:
            root, path = self._root_path
            raise Exception('title was not found in path %s of marker %s'
                            '' % (str(root.id), ':'.join(path)))
        self.__parent = parent
        self.__site = site
        self.__mark_map = mark_map
        self.__marker = marker

        # content(s)
        self.__content = None
        self.__synthetic_cfields = None

    @property
    def _parent(self):
        return self.__parent

    @property
    def _root_path(self):
        paths = []
        parent = self._parent
        while True:
            parent = parent._parent
            if type(parent) is not Marker:
                paths.append(parent.title)
            else:
                break
        root = parent
        paths.reverse()
        return root, tuple(paths)

    @staticmethod
    def title_to_id(title):
        title = title.strip()
        title = title.lower()
        # replace consecutive space chars with single one
        title = ' '.join(title.split())
        # replace no alpha-numeric-dash-underscore chars with a single underscore.
        _id = _mark_title2id_sub_pat.sub('_', title)
        return _id

    @property
    def marker(self):
        return self.__marker

    @property
    def title(self):
        return self.__mark_map['title']

    @property
    def id(self):
        # TODO: create id from title in a better way
        _id = self.__mark_map.get('id', None)
        if _id is None:
            _id = self.title_to_id(self.title)
        return _id

    @property
    def description(self):
        return self.__mark_map.get('description', '')

    @property
    def marks(self):
        return self.__mark_map.get('marks', None)

    @property
    def cfields(self):
        if not self.__marker.is_public:
            raise SynamicMarkerIsNotPublic(
                f'Marker with title {self.__marker.title} is not public and thus the mark title {self.title} cannot '
                f'have cfields'
            )
        if self.__synthetic_cfields is not None:
            sf = self.__synthetic_cfields
        else:
            system_settings = self.__site.system_settings
            content_service = self.__site.get_service('contents')
            url_partition_comp = system_settings['url_partition_comp']
            mark_url_comp = system_settings['mark_url_comp']
            cdoctype = CDocType.GENERATED_HTML_DOCUMENT
            marker_slug = self.__marker.get('slug', None)
            if marker_slug is None:
                marker_slug = self.__marker.id
            curl = self.__site.synamic.router.make_url(
                self.__site, f'/{url_partition_comp}/{mark_url_comp}/{marker_slug}/{self.id}', for_cdoctype=cdoctype
            )
            mimetype = 'text/html'
            sf = synthetic_fields = content_service.make_synthetic_cfields(
                curl,
                cdoctype,
                mimetype,
                cpath=None,
                fields_map=None)
            sf['title'] = self.title
            sf['mark'] = self
            sf['marker'] = self.__marker
            self.__synthetic_cfields = synthetic_fields
            # TODO: create single marker.html template for all and then specific marker id based template?
            # Add such settings to site settings.
        return sf

    @property
    def curl(self):
        return self.cfields.curl

    def __marker_content_renderer(self, site, gen_content):
        site_settings = site.settings
        template_service = site.get_service('templates')
        user_template_name = site_settings['templates.mark']

        html_text_content = template_service.render(
            user_template_name,
            site=site,
            content=gen_content,
            mark=self,
            marker=self.__marker
        )
        return html_text_content

    @property
    def content(self):
        if not self.__marker.is_public:
            raise SynamicMarkerIsNotPublic(
                f'Marker with title {self.__marker.title} is not public and thus the mark title {self.title} cannot '
                f'have content'
            )
        if self.__content is not None:
            content = self.__content
        else:
            content_service = self.__site.get_service('contents')
            content = content_service.build_generated_content(
                self.cfields,
                '',
                source_cpath=None,
                render_callable=self.__marker_content_renderer)
            self.__content = content
        return content

    @property
    def contents(self):
        if not self.__marker.is_public:
            raise SynamicMarkerIsNotPublic(
                f'Marker with title {self.__marker.title} is not public and thus the mark title {self.title} cannot '
                f'have contents'
            )

        if self.__marker.type == 'single':
            query = f'{self.marker.id} == {self.title} :sortby created_on desc'
        else:
            query = f'{self.marker.id} contains {self.title} :sortby created_on desc'
        contents = self.__site.object_manager.query_contents(query)
        return tuple(contents)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.id == other.id and self.marker == other.marker

    def __hash__(self):
        return hash(self.id + self.marker.id)

    def __getattr__(self, key):
        return self.__mark_map.get(key, '')

    def __str__(self):
        return "Mark: %s" % self.title

    def __repr__(self):
        return repr(self.__str__())


class Marker:
    def __init__(self, site, marker_id, marker_type, marker_title, marker_mark_maps, origin_syd):
        assert marker_type in {'single', 'multiple', 'hierarchical'}
        self.__site = site
        self.__id = marker_id
        self.__title = marker_title
        self.__type = marker_type
        # self.__marks_list = mark_maps

        self.__marks_by_title = {}
        self.__marks_by_id = {}

        self.__marks = []
        __marks = []
        self.__process_marks_list(marker_mark_maps, __marks, parent=None)
        self.__origin_syd = origin_syd
        for mark in __marks:
            self.add_mark(mark)

    def make_mark(self, mark_map, parent=None):
        return _Mark(parent, self.__site, mark_map, self)

    def make_mark_by_title(self, title_str, parent=None):
        # TODO: should be empty syd?
        mark_map = {}
        title_str = title_str.strip()
        mark_map['title'] = title_str
        return self.make_mark(mark_map, parent=parent)

    def add_mark(self, mark):
        assert type(mark) is _Mark
        assert mark.id not in self.__marks_by_id, f'{mark.id}'
        self.__marks.append(mark)
        self.__marks_by_title[_Mark.title_to_id(mark.title)] = mark
        self.__marks_by_id[mark.id] = mark

    @property
    def id(self):
        return self.__id

    @property
    def title(self):
        return self.__title

    @property
    def type(self):
        return self.__type

    @property
    def is_public(self):
        public = self.get('is_public', None)
        if public is None or public:
            return True
        return False

    @property
    def is_single(self):
        return self.type == 'single'

    @property
    def is_multiple(self):
        return self.type == 'multiple'

    @property
    def is_hierarchical(self):
        return self.type == 'hierarchical'

    @property
    def marks(self):
        return tuple(self.__marks)

    def get_mark_by_id(self, id, default=None):
        return self.__marks_by_id.get(id, default)

    def get_mark_by_title(self, title, default=None):
        return self.__marks_by_title.get(_Mark.title_to_id(title), default)

    def __process_marks_list(self, mark_maps, res_mark_objs, parent):
        _marks = []
        for mark_map in mark_maps:
            _marks.append(self.make_mark(mark_map, parent))

        if res_mark_objs is not None:
            res_mark_objs.extend(_marks)

        if self.is_hierarchical:
            for _mark in _marks:
                _mark_maps = _mark.marks
                if _mark_maps is not None:
                    self.__process_marks_list(_mark_maps, res_mark_objs, _mark)

    def get(self, key, default=None):
        return self.__origin_syd.get(key, default=default)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.__id)



from abc import abstractmethod, ABC
from .host import HostContract


class SynamicContract(ABC, HostContract):
    @abstractmethod
    def load(self):
        pass

    @property
    @abstractmethod
    def is_loaded(self):
        pass

    @property
    @abstractmethod
    def env(self):
        pass

    @property
    @abstractmethod
    def event_system(self):
        pass

    @property
    @abstractmethod
    def default_data(self):
        pass

    @property
    @abstractmethod
    def object_manager(self):
        pass

    @property
    @abstractmethod
    def sites(self):
        pass

    @property
    @abstractmethod
    def router(self):
        pass



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


def loaded(f):
    def method_wrapper(self, *args, **kwargs):
        assert self.is_loaded, "Class object of the method `%s`'s must be loaded before this method can be called" % f.__qualname__
        return f(self, *args, **kwargs)
    return method_wrapper


def not_loaded(f):
    def method_wrapper(self, *args, **kwargs):
        assert not self.is_loaded, "Must NOT be loaded before this method can be called"
        return f(self, *args, **kwargs)
    return method_wrapper



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

import re
import urllib.parse
from typing import Union
from synamic.core.contracts import CDocType
from synamic.core.standalones.functions.sequence_ops import Sequence
ext_pattern = re.compile(r'.+(\.[a-zA-Z0-9]{1,8})')
# used to check whether an url has an ext so that if it is html cdoctype /index.html is not used.


class ContentUrl:
    @classmethod
    def __str_path_to_comps(cls, path_str):
        comps = []
        for url_comp in re.split(r'[\\/]+', path_str):
            comps.append(url_comp)
        return comps

    @classmethod
    def __sequence_to_comps(cls, path_sequence):
        comps = []
        for url_comp in path_sequence:
            assert isinstance(url_comp, str)
            comps.extend(
                cls.__str_path_to_comps(url_comp)
            )
        return comps

    @classmethod
    def path_to_ccomponents(cls, *url_path_comps: Union[str, list, tuple]) -> tuple:
        res_url_path_comps = []
        for _comps in url_path_comps:
            if isinstance(_comps, str):
                res_url_path_comps.extend(cls.__str_path_to_comps(_comps))
            elif isinstance(_comps, (list, tuple)):
                res_url_path_comps.extend(cls.__sequence_to_comps(_comps))
            else:
                raise Exception('Invalid argument for url component: %s' % str(url_path_comps))

        # ignore empty ones
        _ = []
        for idx, comp in enumerate(res_url_path_comps):
            if idx == 0:  # sparing the first empty string only
                _.append(comp)
            else:
                if comp != '':
                    _.append(comp)
                # else ignore
        res_url_path_comps = _

        # ../../../ recalculation and empty string removing from middle
        _ = []
        for idx, comp in enumerate(res_url_path_comps):
            if comp == '.':
                # just ignore it
                continue
            elif comp == '..':
                #  delete the last comp and add current one (replace the last one)
                if idx == 0:
                    continue
                else:
                    if len(_) >= 1:
                        del _[-1]
            else:
                _.append(comp)
        res_url_path_comps = _

        # re-adding empty string for zero length comps
        if len(res_url_path_comps) == 0:
            res_url_path_comps.append('')

        # comps should begin with empty string ... we can only handle site root absolute url as there is no context for
        # relative url in this function.
        # TODO: DONE: unlike file system path, this last '' should be removed for len(comps) > 1
        # Done by ignoring all empty comps except for the first one -. see above
        # HTML (both gen & non-gen) cdoctype url that does not end with a file extension will have /index.html (take
        # from settings) as real file system url and for browser (client) representation it will end with /.
        # So, for html cdoctype'd urls that have a file extension (place checking so that it does not exceed
        # certain length + the last comp does not start with a dot '.' )
        #  will not end with / for client representation.

        # validating
        for idx, url_comp in enumerate(res_url_path_comps):
            assert not re.match(r'^\s+$', url_comp), "A component of an url cannot be one or all whitespaces"

        return tuple(res_url_path_comps)

    @classmethod
    def path_to_components(cls, *url_path_comps):
        path_comps = cls.path_to_ccomponents(*url_path_comps)
        return Sequence.strip(path_comps, ('',))

    def __init__(self, site, url_path_comps, for_cdoctype=CDocType.UNSPECIFIED):
        """
        append_slash is only for dynamic contents and only when the url_path_comps is being passed as sting (not: list, tuple, content path)
        So, we are not persisting that data
        
        'index.html' in lower case is special throughout the url system - see the codes for extracting more info.
        
        if we indicate ...
        """
        assert isinstance(for_cdoctype, CDocType), \
            f'cURL without a CDocType specified is not acceptable. When you are not going to use the url for final url '\
            f'generation (e.g. for query by url only) you should specify CDocType.UNSPECIFIED\n'\
            f'The type you provided was {type(for_cdoctype)} with the value {for_cdoctype}'
        assert CDocType.is_text(for_cdoctype) or CDocType.is_binary(for_cdoctype) or\
            for_cdoctype == CDocType.UNSPECIFIED

        self.__site = site
        self.__for_cdoctype = for_cdoctype
        # remove space from both end of url (it happens to be only on left.)
        self.__url_path_comps = self.path_to_ccomponents(url_path_comps)

        self.__path_str = None
        self.__path_components_w_site = None
        self.__url_str = None

    def clone(self, for_cdoctype=CDocType.UNSPECIFIED):
        """Besides copying, usable when trying different type of content in router. e.g. static content need path gen"""
        return self.__class__(self.__site, self.__url_path_comps, for_cdoctype=for_cdoctype)

    @property
    def for_site(self):
        return self.__site

    @property
    def for_cdoctype(self):
        return self.__for_cdoctype

    def __join_path_comps_for_url(self, comps):
        # comps must be a result of to_path_components/self.__url_path_comps
        if comps == ('',) or len(comps) == 0:
            path_str = '/'
        else:
            comps = list(comps)
            basename = comps[-1]
            if CDocType.is_html(self.for_cdoctype) and not ext_pattern.match(basename):
                if comps[-1] != '':
                    comps.append('')
            else:
                if comps[-1] == '':
                    del comps[-1]
            path_str = '/'.join(comps)
        return path_str

    def join(self, url_comps: Union[str, list, tuple], for_cdoctype=CDocType.UNSPECIFIED):
        this_comps = self.__url_path_comps
        other_comps = self.path_to_ccomponents(url_comps)
        comps = []
        comps.extend(this_comps)
        comps.extend(other_comps)

        if for_cdoctype is None:
            for_cdoctype = self.__for_cdoctype
        return self.__class__(self.__site, comps, for_cdoctype=for_cdoctype)

    @property
    def path_components(self):
        return self.__url_path_comps

    @property
    def path_components_w_site(self):
        if self.__path_components_w_site is None:
            self.__path_components_w_site = self.path_to_components(
                self.__site.settings.site_base_path,
                self.__url_path_comps
            )
        return self.__path_components_w_site

    @property
    def path_as_str(self):
        assert self.__for_cdoctype is not CDocType.UNSPECIFIED
        if self.__path_str is None:
            self.__path_str = self.__join_path_comps_for_url(self.path_components)
        return self.__path_str

    @property
    def path_as_str_w_site(self):
        assert self.__for_cdoctype is not CDocType.UNSPECIFIED
        if self.__path_str is None:
            self.__path_str = self.__join_path_comps_for_url(self.path_components_w_site)
        return self.__path_str

    @property
    def path_as_str_encoded(self):
        assert self.__for_cdoctype is not CDocType.UNSPECIFIED
        return urllib.parse.quote_plus(self.path_as_str, safe='/:#', encoding='utf-8')

    @property
    def url(self):
        """URL with host name, port, path"""
        assert self.__for_cdoctype is not CDocType.UNSPECIFIED
        if self.__url_str is None:
            ss = self.__site.settings

            path_str = self.path_as_str
            if not path_str.startswith('/'):
                path_str = '/' + path_str

            site_address = ss.site_address
            if site_address.endswith('/'):
                site_address = site_address.rstrip('/')

            url_str = site_address + path_str
            self.__url_str = url_str
        return self.__url_str

    @property
    def url_encoded(self):
        assert self.__for_cdoctype is not CDocType.UNSPECIFIED
        return urllib.parse.quote_plus(self.url, safe='/:#', encoding='utf-8')

    @property
    def to_file_system_path(self):
        # TODO: make it dirfn based
        p = self.path_as_str
        if CDocType.is_html(self.__for_cdoctype):
            index_file_name = self.__site.settings['index_file_name']
            if p.endswith('/'):
                p += index_file_name
            else:
                p += '/' + index_file_name
        else:
            # validation
            # assert not p.endswith('/')
            pass
        return p

    @property
    def to_cpath(self):
        # TODO: make it dirfn based
        return self.__site.path_tree.create_file_cpath(
            self.to_file_system_path
        )

    @property
    def to_cpath_w_site(self):
        # TODO: make it dirfn based
        return self.__site.synamic.path_tree.create_file_cpath(
            self.to_file_system_path
        )

    @property
    def to_dirfn_pair_w_site(self):
        path_comps = self.path_components_w_site
        if len(path_comps) == 0:
            path_comps = ('', )

        if CDocType.is_html(self.__for_cdoctype):
            if ext_pattern.match(path_comps[-1]):
                return path_comps[:-1], path_comps[-1]
            else:
                index_file_name = self.__site.settings['index_file_name']
                return path_comps, index_file_name
        else:
            return path_comps[:-1], path_comps[-1]

    def __str__(self):
        # same as self.path_as_str_w_site, but cannot use that property due to CDocType check issue.
        # keep this method synced with self.path_as_str_w_site
        path_str = self.__join_path_comps_for_url(self.path_components_w_site)
        return f'CURL: {path_str}'

    def __repr__(self):
        return repr(self.__str__())

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return self.path_components_w_site == other.path_components_w_site

    def __hash__(self):
        return hash(self.path_components_w_site)

    @classmethod
    def parse_requested_url(cls, synamic, url_str):
        # python urlparse // bug workaround
        url_str = re.sub(r'(?<!:)/+', '/', url_str)

        parsed_url = urllib.parse.urlparse(url_str)
        url_path = parsed_url.path
        # Unused for now: url_query = parsed_url.query
        # Unused for now: url_fragment = parsed_url.fragment
        path_segments = list(cls.path_to_components(url_path))
        # partition at special url comp
        # TODO: current algo is not respecting site base path. e.g. the root of the site starts at
        # base_path -> example.com/base_path. Fix it.
        site_ids_comps = sorted([site_id.components for site_id in synamic.sites.ids], key=len, reverse=True)
        url_partition_comp = synamic.system_settings['url_partition_comp']

        site_id_components, path_components, special_components = [], [], []
        #  extract out site id.
        for site_id_comps in site_ids_comps:
            if Sequence.startswith(path_segments, site_id_comps):
                site_id_components = path_segments[:len(site_id_comps)]
                path_segments = path_segments[len(site_id_comps):]
                break

        # extract out paginated part
        assert '/' not in url_partition_comp
        assert ' ' not in url_partition_comp
        for idx, segment in enumerate(path_segments):
            if segment == url_partition_comp:
                special_components.extend(
                    path_segments[idx+1:]
                )
                break
            else:
                path_components.append(segment)

        site_id = synamic.sites.make_id('/'.join(site_id_components))

        return site_id, path_components, special_components



"""
This script will be used when the installed synamic shell is not being used and 
the source is in use, for example, during development.
"""
import os
import sys

SYNAMIC_LIBRARY_ROOT = "C:\\...."  # full path to synamic library/src
sys.path.insert(0, SYNAMIC_LIBRARY_ROOT)

from synamic.core.synamic import Synamic
from synamic.entry_points.shell import start_shell


SITE_ROOT_DIR = "test_site"  # put the relative directory name of the site (relative to this script)

site_root = os.path.join(os.path.abspath(os.path.dirname(__file__)), SITE_ROOT_DIR)

print("SITE-ROOT: %s" % site_root)

start_shell(Synamic, site_root)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

import sass
import io
from synamic.core.contracts.content import CDocType
from synamic.core.services.content.generated_content import GeneratedContent


class SCSS_CSSContent(GeneratedContent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.source_cpath is not None
        self.__css_str = None

    def get_stream(self):
        assert self.source_cpath is not None
        if self.__css_str is None:
            scss_fn = self.source_cpath.abs_path
            css_str = sass.compile(filename=scss_fn)
            self.__css_str = css_str
        else:
            css_str = self.__css_str
        css_io = io.BytesIO(css_str.encode('utf-8'))
        return css_io

    def __getitem__(self, item):
        return self.cfields[item]

    def __getattr__(self, item):
        return getattr(self.cfields, item)

    def __str__(self):
        return 'SCSS CSS Generated content for %s' % self.source_cpath.abs_path

    def __repr__(self):
        return repr(self.__str__())



from abc import abstractmethod, ABCMeta


class HostContract(metaclass=ABCMeta):
    """Synamic and Site are hosts currently"""
    @property
    @abstractmethod
    def abs_root_path(self):
        pass

    @property
    @abstractmethod
    def path_tree(self):
        pass

    @property
    @abstractmethod
    def default_data(self):
        pass



# lots of thoughts needed.

"""
-------------------------------------------------------------------------
--------- comments -------------
// is a single line comment.
# is also a single line comment.
Where comments can appear:
1. Inside a block: on an independent line that does not contain scalar data
2. Inside a block list: on an independent line that does not contain scalar data
3. After the starting curly brace/bracket of block/list.
4. On an empty line outside of blocks/block lists. That is on the independent lines of file. Be noted that, a file is considered block and it
    Contains key value pairs.

Where it cannot appear:
Within or at the end of scalar data: string, number, date
Within or at the end of linear list: (a, t, 5)

Comments may be preceded by spaces and/or tabs.

\# is a literal hash. Hash inside strings are also literal hash.
\// is also literal two forward slash.
--------------------------------

--------- import/include --------
Other data files can also be imported/included with !include <relative file name with or without ext>

!include must be on it's own line.

--------- load model ------------
Models can be specified by the system
---------------------------------

--------- interpolation ---------
Interpolation happens in the second pass. So, after everything is loaded, the non-recursive keys can be found
easily.

{{ key }}

{{ key.subkey }}

key can be found in the current data file or from an included file.

To put literal {{ then just precede it with a single backward slash. \{{
This is not needed for single curly brace.

-------- escaping ----------------
Escaping is done through backward slash. Escaping are relaxed and special in synamic data file.
To escape a backward slash escape it with another one.
Escaping does not take place in each and every thing.
----------------------------------
---------- CALCEL, don't want to keep type hinting, need to make things clean for end user | data __document_types ----------
* Data type inforcement *
If you want to force type on a data then you can use the following key suffixes.
!n -> number
!i -> integer
!f -> float
!d -> date
!t -> time
!dt -> date-time
!s -> string
!l -> list
!b -> boolean

* Other suffixes can also be defined by the system with models *

1) >> Number: Priority Top (1) <<
-> 1 <- is a number.
-> 2.5 <- is a number.
# 1 & 1.0 are equal.

2) >> Date, Time & Datetime: Priority 2 <<
If something falls in the format of date, time or datetime that is considered is date.
The default formats are.

If you want a date like format to be treated like string then use quoted string.

Date: <4 digit year>:<1 or 2 digit month within the range of 1-12>:<1 or 2 digit date within the range defined by month>
Time: <1 or 2 digit hour within range of 0-23 or 0-12 if AM/PM present>:<1 or 2 digit minute 1-59>:<Optional: 1 or 2 digit seconds>:<Optional: AM/PM - case insensitive>
Datetime: Combination of date & time separated by space character(s) - spaces/tabs.

3) >> String: Priority 3 <<
-> Strings 1 (no quotation): strings does not need any quotation mark. No special/escape characters - everything is literal char.
key: this is a string value, leading and trailing spaces are ignored.

-> String 2 (single quotation): if the value part starts with a single quotation then it is considered single
quoted string. a quotation inside it must be escaped with backward slash. 

4) >> List: Priority 4
* If a key ends with "!l" or "!L" then it will be considered a list. So, there is no need to to delimit
list elements with [ & ]

- List elements are separated by commas. You cannot enforce type on list elements, they must be inferred.
- if you want special treatment then use model based parsing.

- List elements can contain numbers, single line strings, date, time, date-time, boolean data __document_types only.
- Strings must be single line strings. You can use multiline strings with triple quoting.

4) >> Other data: Priority 3 <<
If a key is mentioned in an associated model then the string is parsed.

 

--------------------------------
-------------------------------------------------------------------------


!include values-dev.txt
!include_if_exists values-super-dev2.txt

key1 : value
// data type specification cancelled: key10 !s the colon after the key is optional.
// data type specification cancelled: key11 !l : this is, a typed, key, with type list
key2 {
        multilevel: value 2
}

# multiline text: key must be suffixed with ~, in this case there must not be any type specifier.
# as described above, the colon is optional. But a multiline text must be enclosed in curly braces.
key3~: {

}

# if the there is no text after { then this empty line is not included in the output.
# if it contains some text then the preceding spaces are also considered to be included in the calculation of indentaion.

key30~ {
         Indentation does not start here.
  Indentation starts here.
  Tab normalization happens (no replace - just counted virtually) for counting indentation. One tab is considered
     as 4 spaces.
}

# if you do not want any indentation detection happen and take all the text literally, then you must use 
# double tilde.

key31~~ : {
   All preceded spaces are kept
Intact, no matter at how many level the key is nested.
}

# Nested values

key4: {  
    k1: {
         k11: value
    }
    k2: single line
    multiline~ {
    }
}
"""

import re
import enum
import pprint
import datetime
import numbers
from synamic.core.standalones.functions.date_time import DtPatterns, parse_date, parse_time, parse_datetime
from synamic.exceptions import SynamicSydParseError, get_source_snippet_from_text


# patterns
class _Patterns:
    key_pattern = re.compile(r'''^[ \t]*
            (?P<key>[a-zA-Z_]+[a-zA-Z0-9_]*) # key
            [ \t]*
            (?P<is_multiline>~{1,3})? # is multi line
            [ \t]*
            # data type specification cancelled (?P<type>![a-zA-Z0-9_]+)?[ \t]*    # enforced format. If it is specified in multiline code and no model is specified
                                   # then that format is ignored.
            :? # colon is optional
            (?={|\[|\s|\Z)
            ''', re.X)

    number = re.compile(r'^[ \t]*(?P<number>[+-]?[0-9]+(\.[0-9]+)?)[ \t]*$')
    date = DtPatterns.date
    time = DtPatterns.time
    datetime = DtPatterns.datetime

    inline_list_separator = re.compile(r'(?<!\\),')

    multiline_spaces = re.compile(r'^[ \t]*')

    inline_list = re.compile(r'^[ \t]*\((?P<content>.*?)\)[ \t]*$')

    interpolation_identifier = re.compile(r'(?<!\$)\$\{[ \t]*(?P<identifier>'
                                          r'[a-z_][a-z0-9_]*)[ \t]*\}', re.I)  # escape $ with another - that simple


class SydParseError:
    def __init__(self, line, line_no, invalid_data, col_no=None, msg=''):
        self.line = line
        self.line_no = line_no
        self.invalid_data = invalid_data
        self.col_no = col_no
        self.msg = msg


@enum.unique
class SydDataType(enum.Enum):
    number = 1
    date = 2
    time = 3
    datetime = 4
    string = 5

    # data structures
    # block = 6 # map
    # block_list = 7
    # inline_list = 8


syd_to_py_types = {
    SydDataType.time: (datetime.time, ),
    SydDataType.date: (datetime.date, ),
    SydDataType.datetime: (datetime.datetime, ),
    SydDataType.number: (numbers.Number, int, float),
    SydDataType.string: (str, )
}

py_to_syd_types = {}
for syd_type, py_types in syd_to_py_types.items():
    for py_type in py_types:
        py_to_syd_types[py_type] = syd_type


class __SydData:
    @property
    def is_scalar(self):
        return type(self) is SydScalar

    @property
    def is_container(self):
        return type(self) is SydContainer

    def syd_set_parent(self, p):
        raise NotImplemented

    def set_converter(self, c):
        raise NotImplemented

    def set_converted(self, v):
        raise NotImplemented

    @property
    def converter(self):
        raise NotImplemented


class SydScalar(__SydData):
    def __init__(self, key, value, datatype, parent_container=None, converter=None, converted_value=None):
        if key is not None:
            assert '.' not in key, f'Key {key} is invalid where value is {value}'
        assert isinstance(value, syd_to_py_types[datatype])
        self.__key = key
        self.__value = value
        self.__datatype = datatype
        self.__converter = converter
        self.__converted_value = converted_value
        self.__parent_container = None

        self.__cached_interpolated_str = None
        self.syd_set_parent(parent_container)

    def clone(self, parent_container=None, converter=None, converted_value=None):
        if converter is None:
            converter = self.__converter
        if converted_value is None:
            converted_value = self.__converted_value
        return self.__class__(self.__key, self.__value, self.__datatype, parent_container=parent_container, converter=converter, converted_value=converted_value)

    @property
    def key(self):
        return self.__key

    @property
    def value(self):
        if self.__converted_value is not None:
            value = self.__converted_value
        else:
            if not callable(self.__converter):
                value = self.__value
            else:
                value = self.__converter(self.__value)
                assert value is not None
                self.__value = value
        return value

    @property
    def value_origin(self):
        return self.__value

    def set_converted(self, value):
        assert value is not None
        self.__converted_value = value

    @property
    def type(self):
        return self.__datatype

    def syd_set_parent(self, p):
        assert type(p) in (SydContainer, type(None))
        self.__parent_container = p

        # process interpolation
        if p is not None:
            value = self.__value
            # process interpolation for string
            if self.__datatype is SydDataType.string:
                if self.__cached_interpolated_str is not None:
                    value = self.__cached_interpolated_str
                else:
                    value = self.__cached_interpolated_str = _Patterns.interpolation_identifier.sub(self.__interpolation_replacer,
                                                                                                    value)
            self.__value = value

    def set_converter(self, converter):
        assert not callable(self.__converter)
        self.__converter = converter

    @property
    def converter(self):
        return self.__converter

    def __str__(self):
        if self.key is not None:
            return "%s => %s" % (self.key, str(self.value))
        else:
            return str(self.value)

    def __repr__(self):
        return repr(self.__str__())

    def __interpolation_replacer(self, match):
        identifier = match.group('identifier')
        assert self.__key != identifier, 'Key is being interpolated recursively: %s' % identifier
        if self.__parent_container is not None:
            v = self.__parent_container.get(identifier, '')
            assert type(v) in (int, float, str), 'Only number and string data can be interpolated in' \
                                                 'to a string. Data with' \
                                                 ' kye `%s` is of type `%s`' % (self.key, str(type(v)))
            v = str(v)  # convert it in case it is a number.
            return v
        else:
            raise Exception('Parent for key `%s` is set to None' % str(self.__key))


class SydContainer(__SydData):
    def __init__(self, key, is_list=False, parent_container=None, converter=None, converted_value=None, read_only=False):
        self.__key = key
        self.__is_list = is_list
        self.__map = {}
        self.__list = []
        self.__parent_container = None  # parent_container from init will be set through a method below.
        self.__converter = converter
        self.__converted_value = converted_value
        self.__read_only = read_only

        self.syd_set_parent(parent_container)

    def clone(self, parent_container=None, converter=None, converted_value=None, read_only=False):
        if converter is None:
            converter = self.__converter
        if converted_value is None:
            converted_value = self.__converted_value
        cln = self.__class__(self.__key, self.__is_list, parent_container=parent_container, converter=converter, converted_value=converted_value, read_only=read_only)
        for e in self.__list:
            cln.add(e.clone())
        return cln
    copy = clone

    @property
    def is_root(self):
        return self.__parent_container is None or self.__key == '__root__' or self.__key is None

    # methods prefixed with syd_ are considered private functions or functions should not be used by users.
    def syd_clone_original_contents(self):
        l = []
        for e in self.__list:
            l.append(e.clone())
        return tuple(l)

    def syd_get_original(self, key, multi=False):
        assert isinstance(key, (int, str, list, tuple)), 'Only integer and string keys are accepted, you provide key of type: %s' % str(
            type(key))
        key = int(key) if type(key) is str and key.isdigit() else key

        if type(key) is int:
            assert self.is_list, "Index %d provided for a block collection. Use numeric index only for list collections" % key
            try:
                d = self.__list[key]
                if multi:
                    d = (d, )

            except IndexError:
                raise IndexError('%d does not exist')
        else:
            if isinstance(key, (list, tuple)):
                keys = key
            else:
                keys = key.split('.')
            try:
                first_key = keys[0]
                cont_tuple = tuple(t[1] for t in self.__map[first_key])
                keys = keys[1:]
                last_idx = len(keys) - 1
                for idx, _key in enumerate(keys):
                    if last_idx == idx:
                        if not cont_tuple[-1].is_container:
                            # print('Found value is not a block - it is a scalar')
                            raise KeyError('Found value is not a block - it is a scalar')
                    _ = cont_tuple[-1].syd_get_original(_key, multi=multi)

                    if multi:
                        cont_tuple = _
                    else:
                        cont_tuple = (_, )

                if multi:
                    d = cont_tuple
                else:
                    d = cont_tuple[-1]
            except KeyError:
                raise KeyError('Key `%s` was not found' % key)
        return d

    # def copy_

    @property
    def is_list(self):
        return self.__is_list

    @property
    def key(self):
        return self.__key

    def keys(self):
        l = list()
        for e in self.__list:
            if e.key is not None:
                l.append(e.key)
        return tuple(l)

    def values(self):
        """Values are not converted!!!"""
        l = list()
        for e in self.__list:
            value = e.value
            l.append(value)
        return tuple(l)

    def items(self):
        """Values are not converted!!!"""
        l = []
        i = 0
        for e in self.__list:
            value = e.value
            if self.is_list:
                l.append((i, value))
            else:
                l.append((e.key, value))
            i += 1
        return tuple(l)

    def __getitem__(self, key):
        value = self.get(key, default=None, multi=False)
        if value is None:
            raise KeyError('Key `%s` was not found' % key)
        return value

    def set_converter(self, converter):
        assert not callable(self.__converter)
        assert not self.is_root
        self.__converter = converter

    def set_converter_for(self, key, converter):
        syds = self.syd_get_original(key, multi=True)
        for syd in syds:
            syd.set_converter(converter)

    def get(self, key, default=None, multi=False):
        try:
            value = self.syd_get_original(key, multi=multi)
            if not multi:
                value = value.value
            else:
                values = value
                _ = []
                for v in values:
                    _.append(v.value)
                value = tuple(_)
        # except (KeyError, IndexError, AssertionError):
        except (KeyError, IndexError):
            value = default
        return value

    def add(self, syd):
        assert not self.__read_only
        self.__list.append(syd)
        if not self.is_list:
            idx = len(self.__list) - 1
            key = syd.key
            if key not in self.__map:
                l = []
                self.__map[key] = l
            else:
                l = self.__map[key]
            l.append((idx, syd))
        syd.syd_set_parent(self)

    def set(self, key, py_value):
        """
        Sets py value as converted to existing key
        BUT for non existing value it creates scalar of only allowed type.
        """
        assert not self.__read_only
        try:
            syd = self.syd_get_original(key, multi=False)
            syd.set_converted(py_value)
        except KeyError:
            if '.' not in key:
                parent_container = self
            else:
                comps = key.split('.')[:-1]
                parent_container = self.syd_get_original(comps, multi=False)
                assert parent_container.is_container

            syd_scalar = SydScalar(
                key,
                py_value,
                py_to_syd_types[type(py_value)],
                converter=None,
                converted_value=py_value
            )
            parent_container.add(syd_scalar)

    def __setitem__(self, key, value):
        assert not self.__read_only
        self.set(key, value)

    def __remove__from__self(self, key):
        key = int(key) if type(key) is str and key.isdigit() else key
        if self.is_list:
            del self.__list[key]
        else:
            idx_syd_l = self.__map[key]
            idx_syd_l.reverse()

            del self.__map[key]
            for idx, syd in idx_syd_l:
                del self.__list[idx]

            new_map = type(self.__map)()
            for idx, e in enumerate(self.__list):
                if e.key not in new_map:
                    l = []
                    new_map[e.key] = l  # Here was that deadly bug
                else:
                    l = new_map[e.key]
                l.append((idx, e))
            self.__map = new_map

    def __delitem__(self, key):
        assert not self.__read_only
        assert type(key) in (int, str), 'Only integer and string keys are accepted, you provide key of type: %s' % str(
            type(key))
        key = str(key)
        ks = key.split('.')
        key2del = ks[-1]
        ks = ks[:-1]

        try:
            if len(ks) == 0:
                cont = self
            else:
                k = ks[0]
                ks = ks[1:]
                cont = self.syd_get_original(k)
                for k in ks:
                    cont = cont._get_original(k)
            cont.__remove__from__self(key2del)
        except (KeyError, IndexError):
            raise KeyError('Key/Index `%s` was not found' % key)

    def __contains__(self, key_value):
        """This method only checks for existence of a key - not value, even if this object is a list"""
        if self.is_list:
            value = key_value
            return value in self.__list
        else:
            key = key_value
            res = self.get(key, None)
            return False if res is None else True

    def key_exists(self, key):
        return key in self

    @property
    def value(self):
        if self.__converted_value is not None:
            value = self.__converted_value
        else:
            if not callable(self.__converter):
                value = self.value_origin
            else:
                value = self.__converter(self.value_origin)
                assert value is not None
                self.__converted_value = value
        return value

    @property
    def value_origin(self):
        if self.is_list:
            c = []
            for a in self.__list:
                c.append(a.value)
            c = tuple(c)
        else:
            # c = collections.OrderedDict()
            c = {}
            for a in self.__list:
                c[a.key] = a.value
        return c

    def set_converted(self, value):
        assert value is not None
        assert not self.is_root
        self.__converted_value = value

    def new(self, *others):
        assert not self.is_list, 'Cannot create new from list, it must be a block'
        self_clone = self.clone()
        for other in others:
            assert not other.is_list, 'Cannot create new with list, need block'
            for oe in other.syd_clone_original_contents():
                if oe.key in self_clone:
                    del self_clone[oe.key]
                self_clone.add(oe)
        return self_clone

    def __str__(self):
        l = []
        for e in self.__list:
            l.append(str(e))
        if self.is_list:
            data_text = ", ".join(l)
        else:
            data_text = "\n".join(l)
        if self.key is not None:
            return str(self.key) + " { %s }" % data_text
        else:
            res = "{ %s }" % data_text
            return res

    def __repr__(self):
        return repr(self.__str__())

    def syd_set_parent(self, p):
        assert type(p) in (type(None), self.__class__), 'Invalid type: %s' % str(type(p))
        self.__parent_container = p

    @property
    def converter(self):
        return self.__converter


@enum.unique
class _ParseState(enum.Enum):
    processing_block = 1
    processing_list = 1.1
    processing_block_string = 3  # multiline string
    processing_inline = 2
    processing_inline_list = 2.2
    processing_nothing = 0

    def __str__(self):
        return str(self.name)

    def __repr__(self):
        return str(self)


class SydParser:
    def __init__(self, text, debug=False):
        self.__text = text
        self.__debug = debug
        self.__tree = SydContainer('__root__')
        self.__error = None
        self.__lines = self.__text.splitlines()
        self.__current_line_no = 0

        self.__parse_states = []
        self.__container_stack = [self.__tree]

    def __dprint(self, *args, **kwargs):
        if self.__debug:
            print(*args, **kwargs)

    @property
    def __fmt_error_msg(self):
        line = self.__lines[self.__current_line_no - 1]
        return f"Line no: {self.__current_line_no}\n"\
            f"Line: {line}\nCurrent Processing States Stack: {', '.join(str(state) for state in self.__parse_states)}"

    def __enter_state(self, state: _ParseState):
        self.__parse_states.append(state)

    def __leave_state(self, state: _ParseState):
        assert type(state) is _ParseState
        assert self.__parse_states[-1] == state
        self.__parse_states.pop()

    def __leave_last_state(self):
        self.__parse_states.pop()

    @property
    def __current_state(self):
        if len(self.__parse_states) == 0:
            return _ParseState.processing_nothing
        return self.__parse_states[-1]

    @property
    def __current_level(self):
        return len(self.__parse_states)

    def parse(self):  # (-1)
        if len(self.__lines) != 0:
            assert self.__current_line_no == 0
            self.__enter_state(_ParseState.processing_block)
            self.__process_block('__root__')  # root doc is considered a single block.
            self.__leave_state(_ParseState.processing_block)
        return self.__tree

    def __process_block(self, key, block_type=None):  # (0)
        """This loop is used at the root and at any nested level (root from that perspective)"""
        assert self.__current_state in (_ParseState.processing_block, _ParseState.processing_list)
        if key != '__root__':
            if self.__current_state == _ParseState.processing_list:
                is_list = True
            else:
                is_list = False
            parent_container = SydContainer(key, is_list)
            root_container = self.__container_stack[-1]
            root_container.add(parent_container)
            self.__container_stack.append(parent_container)

        while self.__current_line_no < len(self.__lines):
            # this top level while lool will process only the top level lines/key-values
            self.__current_line_no += 1
            line = self.__lines[self.__current_line_no - 1]
            self.__dprint("Line %s: %s" % (self.__current_line_no, line))

            # ignore empty line: empty lines are ignored only when they are at the top level (no matter how indented)
            stripped_line = line.strip()
            if stripped_line == '':
                continue
            # ignore comment
            elif self.__is_comment(stripped_line):
                continue
            # leave block
            is_block_end, block_end_type = self.__block_end(line)
            if is_block_end:
                if block_type == '{':
                    if block_end_type == '}':
                        # self.current_line_no += 1
                        break
                elif block_type == '[':
                    if block_end_type == ']':
                        # self.current_line_no += 1
                        break
            self.__process_block_items()
        del self.__container_stack[-1]

    def __process_key_value(self):
        pass

    def __process_value(self):
        pass

    def __process_block_items(self):  # (1)
        line = self.__lines[self.__current_line_no - 1]
        #  text = line
        #  extract key
        if self.__current_state in (_ParseState.processing_list, _ParseState.processing_block_string):
            key = None
            end_pos = 0
            is_multiline = False
            multiline_token = None
        else:
            key_match = _Patterns.key_pattern.match(line)
            if key_match:
                key = key_match.group('key')
                end_pos = key_match.end()
                multiline_token = key_match.group('is_multiline')
                if multiline_token:
                    is_multiline = True
                else:
                    is_multiline = False
            else:
                self.__dprint("error in line: %s" % line)
                err = SynamicSydParseError(
                    f'Syd Parsing key match error:\n{self.__fmt_error_msg}\n'
                    f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                )
                raise err

        # self.__dprint("\nKEY: %s" % str(key))
        # self.__dprint("\nLINE END: %s" % line[end_pos:])
        # process multi line block
        # multi line string processing

        line_end = line[end_pos:].lstrip()
        next_line = self.__next_line
        is_block_start, block_type = self.__block_start(line_end, next_line)
        inline_list_match = _Patterns.inline_list.match(line_end)
        if is_multiline:
            # end_pos += multiline_match.end()
            self.__enter_state(_ParseState.processing_block_string)
            self.__process_datum_ml_string(end_pos, key, multiline_token)
            self.__leave_state(_ParseState.processing_block_string)
        # so it is not a multi line stuff
        # is it nested stuff?
        elif is_block_start:
            if block_type == '{':
                self.__enter_state(_ParseState.processing_block)
                # self.current_line_no += 1
                self.__process_block(key, block_type)
                #
                self.__leave_state(_ParseState.processing_block)
            elif block_type == '[':
                self.__enter_state(_ParseState.processing_list)
                self.__process_block(key, block_type)
                self.__leave_state(_ParseState.processing_list)
            else:
                err = SynamicSydParseError(
                    f'Syd Parsing error - Something is horribly wrong:\n{self.__fmt_error_msg}\n'
                    f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                )
                raise err
        # inline list
        elif inline_list_match:
            content = inline_list_match.group('content')
            cnt = SydContainer(key, True)
            self.__container_stack[-1].add(cnt)
            self.__container_stack.append(cnt)
            self.__enter_state(_ParseState.processing_inline_list)
            parts = self.__process_datum_inline(content, None, True)
            for part in parts:
                cnt.add(part)
            self.__leave_state(_ParseState.processing_inline_list)
            del self.__container_stack[-1]
        elif self.__is_inline(line_end, next_line):
            self.__enter_state(_ParseState.processing_inline)
            data = self.__process_datum_inline(line_end, key)
            self.__container_stack[-1].add(data)
            # must not increment as it it is inline: X self.current_line_no += 1
            self.__leave_state(_ParseState.processing_inline)
        else:
            err = SynamicSydParseError(
                f'Syd Parsing error - could not parse:\n{self.__fmt_error_msg}\n'
                f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
            )
            raise err

    def __process_datum_ml_string(self, end_pos, key, multiline_token):  # (1.1)
        line = self.__lines[self.__current_line_no - 1]
        line_end = line[end_pos:].lstrip()
        if not line_end.startswith('{'):
            err = SynamicSydParseError(
                f'Syd Parsing error - ...:\n{self.__fmt_error_msg}\n'
                f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
            )
            raise err
        else:
            lines = []
            line_end = line_end[1:]

            # test & ignore first line if valid code follows (starts with {)
            if line_end.strip() == '' or self.__is_comment(line_end):
                pass  # ignore empty line or comment for the first line.
            else:
                err = SynamicSydParseError(
                    f'Syd Parsing error - Multiline string starting line cannot contain data (only comment or blank)'
                    f':\n{self.__fmt_error_msg}\n'
                    f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                )
                raise err

            # now collect lines until you get a } on a single line.
            # if a line contains only one } and you want to make that literal then escape it with \
            # extract lines
            while True:
                self.__current_line_no += 1
                line = self.__lines[self.__current_line_no - 1]
                is_block_end, end_char = self.__block_end(line)
                if is_block_end and end_char == '}':
                    break
                if line.strip(' \t') == r'\}':
                    # unescape starting curly brace.
                    line = line.replace(r'\}', '}')

                lines.append(line)
                if self.__current_line_no == len(self.__lines):
                    err = SynamicSydParseError(
                        f'Syd Parsing error - End of text but no end to the newline found'
                        f':\n{self.__fmt_error_msg}\n'
                        f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                    )
                    raise err

                    # process lines
                    # > if is_multiline == '~~'
                    # blah blah blah
            if multiline_token == '~':
                processed_lines = []
                ind_sizes = []
                for line in lines:
                    match = _Patterns.multiline_spaces.match(line)
                    spaces = match.group() if match.group() is not None else ''
                    size = 0
                    for space in spaces:
                        if space == ' ':
                            size += 1
                        elif space == '\t':
                            size += 4
                    ind_sizes.append(size)
                if len(ind_sizes) > 0:
                    ind_size = min(ind_sizes)
                else:
                    ind_size = 0

                # process lines.
                for line in lines:
                    idx = 0
                    ind_limit = 0
                    for char in line:
                        while ind_limit < ind_size:
                            if char == ' ':
                                ind_limit += 1
                            elif char == '\t':
                                ind_limit += 4
                            else:
                                break
                            idx += 1
                    res_line = line[idx:]
                    processed_lines.append(res_line)
            elif multiline_token == '~~':
                processed_lines = lines
            else:
                assert multiline_token == '~~~'
                processed_lines = lines
                # TODO: process for escape chars & special sequences.

            data = SydScalar(key, '\n'.join(processed_lines), SydDataType.string)
            self.__container_stack[-1].add(data)

    def __process_datum_inline(self, text, key, processing_inline_list=False):
        try:
            return self.convert_to_scalar_values(text, key, processing_inline_list=processing_inline_list)
        except ValueError as ve:
            err = SynamicSydParseError(
                f'Syd Parsing error - value error ({ve.args[0]}):\n{self.__fmt_error_msg}\n'
                f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
            )
            raise err

    @staticmethod
    def __extract_key(text):
        match = _Patterns.key_pattern.match(text)
        if not match:
            return None
        key = match.group('key')
        return key

    @staticmethod
    def __is_comment(line):
        """ Is this LINE a comment"""
        # print(repr(line.splitlines()))
        assert len(line.splitlines()) == 1
        line = line.lstrip(' \t')
        if line.startswith('#') or line.startswith('//'):
            return True
        return False

    def __is_inline(self, text, next_line):
        if next_line is None:
            return True
        is_block_start, _ = self.__block_start(text, next_line)
        if is_block_start:
            return False
        else:
            return True

    def __block_start(self, text, next_line):
        """A multiline is a block. A nested group is also a block"""
        lstripped_text = text.lstrip(' \t')
        if lstripped_text.startswith('{') or \
                lstripped_text.startswith('['):
            # is the rest comment
            text_end = lstripped_text[1:]
            text_end_stripped = text_end.strip(' \t')
            if (text_end_stripped == '' or
                    self.__is_comment(text_end)) and \
                    next_line is not None:
                # so this is a block
                return True, lstripped_text[0]
            else:
                err = SynamicSydParseError(
                    f'Syd Parsing error - Something is horribly wrong!:\n{self.__fmt_error_msg}\n'
                    f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                )
                raise err
        return False, None

    @staticmethod
    def __block_end(line):
        stripped_line = line.strip(' \t')
        if stripped_line == '}':
            return True, '}'
        elif stripped_line == ']':
            return True, ']'
        return False, None

    @property
    def __next_line(self):
        next_line = None
        if self.__current_line_no < len(self.__lines):  # current line is the last line:
            next_line = self.__lines[self.__current_line_no]
        return next_line

    #  InlineDataProcessors
    single_quoted_string_stop = re.compile(r"(?<!\\)'")
    double_quoted_string_stop = re.compile(r'(?<!\\)"')

    @classmethod
    def __single_q_string(cls, line):
        return cls.__quoted_string(line, "'")

    @classmethod
    def __double_q_string(cls, line):
        return cls.__quoted_string(line, '"')

    @classmethod
    def __quoted_string(cls, line, q):
        assert q in ("'", '"')
        res = ''
        end_pos = 0

        if q == '"':
            m = cls.double_quoted_string_stop.search(line)
        else:
            m = cls.single_quoted_string_stop.search(line)

        if not m:
            res += line
            end_pos = len(line)
        else:
            start = line[:m.end() - 1]
            res += start
            end_pos += m.end()
        res = res.replace(r"\%s" % q, "%s" % q)
        end = line[end_pos:]
        return res, end, end_pos

    @classmethod
    def convert_to_scalar_values(cls, text, key, processing_inline_list=False):
        data_list = []
        text = text.strip(' \t\r\n')

        while text:
            # single quoted string
            if text.startswith(("'", '"')):
                if text.startswith("'"):  # and text.endswith("'"):
                    content = text[1:]
                    res, text, _ = cls.__single_q_string(content)
                    data = SydScalar(key, res, SydDataType.string)
                # double quoted string
                else:  # text.startswith('"'):  # and text.endswith('"'):
                    content = text[1:]
                    res, text, _ = cls.__double_q_string(content)
                    res = res.replace(r'\n', '\n')
                    res = res.replace(r'\r', '\r')
                    res = res.replace(r'\t', '\t')
                    data = SydScalar(key, res, SydDataType.string)

                if processing_inline_list:
                    next_comma_match = _Patterns.inline_list_separator.search(text)
                    if next_comma_match:
                        next_stop = next_comma_match.end()
                        text = text[next_stop:]
                else:
                    if text.strip() != '':
                        err = SynamicSydParseError(
                            f'Syd Parsing error - We are not processing inline list, but still there are some data left'
                            f' after converting one:\n{self.__fmt_error_msg}\n'
                            f'Details:\n{get_source_snippet_from_text(self.__text, self.__current_line_no, limit=10)}'
                        )
                        raise err
            else:
                orig_end = text
                next_stop = len(text)
                data_part = text[:next_stop]
                text = text[next_stop:]
                if processing_inline_list:
                    next_comma_match = _Patterns.inline_list_separator.search(orig_end)
                    if next_comma_match:
                        next_stop = next_comma_match.end()
                        data_part = orig_end[:next_stop - 1]
                        text = orig_end[next_stop:]
                data_part = data_part.strip()

                number_match = _Patterns.number.match(data_part)
                datetime_match = _Patterns.datetime.match(data_part)
                date_match = _Patterns.date.match(data_part)
                time_match = _Patterns.time.match(data_part)

                # number
                if number_match:
                    num_str = number_match.group('number')
                    if num_str.startswith(('+', '-')):
                        sign = num_str[0]
                        num_str = num_str[0:]
                        sign_mul = 1 if sign == '+' else -1
                    else:
                        sign_mul = 1
                    if data_part.isdigit():
                        number = int(data_part)
                    else:
                        number = float(data_part)
                    number = number * sign_mul
                    data = SydScalar(key, number, SydDataType.number)
                # date-time match
                elif datetime_match:
                    dt_instance = parse_datetime(datetime_match)
                    data = SydScalar(key, dt_instance, SydDataType.datetime)
                # date match
                elif date_match:
                    date_instance = parse_date(date_match)
                    data = SydScalar(key, date_instance, SydDataType.date)

                # time match
                elif time_match:
                    time_instance = parse_time(time_match)
                    data = SydScalar(key, time_instance, SydDataType.time)
                # bare string : last resort
                else:
                    bare_string = data_part
                    if len(bare_string) > 1:
                        if bare_string[0:2] in (r'\(', r'\{', r'\['):
                            bare_string = bare_string[2:]
                    bare_string = bare_string.strip()
                    data = SydScalar(key, bare_string, SydDataType.string)

            data_list.append(data)
            text = text.strip()

        if len(data_list) == 0:
            # bare string
            data_list.append(
                SydScalar(key, '', SydDataType.string)
            )

        if processing_inline_list:
            return data_list
        else:
            return data_list[0]

    @classmethod
    def covert_one_value(cls, text):
        syd_scalar = cls.convert_to_scalar_values(text, '__null__', processing_inline_list=False)
        return syd_scalar.value


if __name__ == '__main__':
    text = """
    a : {
        what are you doing
    }
    ml ~ {
        my
         text
    }
    ml2 t
    k : v
    m : n
    ikndf /lj
    key1{
        key_n1 : {
            k: 1
            l m
            o {
                p ; l:9;
            } 
        }
    }
    key2: key2data
    list: (uj, 87   ,     ll)
    
        listn: ((8\, 7))
    str2qu: "double quoted string"
    str1qu: 'single quoted string'
    num: 5.4
    lis1[#
    my name
        
        yes
        ali
        {
            k: v
        }
        ("l" ,           7')
        (    7,   8)
        "JJJ Hye"
        ~{
        ml srting in list
            got it?
        }
    ]
    lis1{
        k m
    }
    """
    text2 = """
    max1: 5
    lis1: xx
    """
    s = SydParser(text, None, debug=True)
    s2 = SydParser(text2, None, debug=True)
    print("--- parsing ----")
    tree = s.parse()
    tree2 = s2.parse()
    ntree = tree.new(tree2)
    print("--- the tree ---")
    print(tree)
    # print(SpecialTokens.tokens)
    print('--- pyvalue ---\n')
    pprint.pprint(tree.value)
    # print('--- is collection ---\n')
    # pprint.pprint(tree.is_collection)
    # print('--- keys ---\n')
    # pprint.pprint(tree.keys())
    print('--- get multi ---\n')
    pprint.pprint(tree.get('lis1', multi=True))
    print('--- ... in ... ---\n')
    pprint.pprint('lis1.3' in tree)

    print(' --------- tree.new tree2')
    print(ntree)





import types
from collections import defaultdict, OrderedDict
from synamic.core.services.content.content_splitter import content_splitter
from synamic.core.parsing_systems.model_parser import ModelParser
from synamic.core.parsing_systems.curlybrace_parser import SydParser
from synamic.core.standalones.functions.decorators import loaded, not_loaded
from synamic.core.contracts import CDocType
from .query import QueryNode, SimpleQueryParser
from synamic.core.parsing_systems.getc_parser import parse_getc
from synamic.core.standalones.functions.sequence_ops import Sequence
from synamic.exceptions import (
    SynamicGetCParsingError,
    SynamicGetCError,
    SynamicMarkNotFound,
    SynamicMarkerNotFound,
    SynamicSydParseError,
    SynamicErrors,
    SynamicFSError,
    SynamicSiteNotFound,
    SynamicUserNotFound
)
from synamic import Nil
from .init_manager import InitManager


class ObjectManager:
    def __init__(self, synamic):
        self.__synamic = synamic

        self.__site_object_managers = OrderedDict({})
        self.__site_settings = defaultdict(dict)

        self.__cache = self.__Cache(self.__synamic)

        self.__is_loaded = False

    def get_manager_for_site(self, site):
        if site.id not in self.__site_object_managers:
            om_4_site = self.__ObjectManagerForSite(site, self)
            self.__site_object_managers[site.id] = om_4_site
        else:
            om_4_site = self.__site_object_managers[site.id]
        return om_4_site

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        self.__is_loaded = True

    def __reload_for__(self, site):
        self.__cache.clear_cache(site)
        self.__load_for__(site)

    def __load_for__(self, site):
        self.__cache_markers(site)
        self.__cache_users(site)
        self.__cache_marked_cfields(site)
        self.__cache_menus(site)
        self.__cache_data(site)
        self.__cache_pre_processed_contents(site)

    def __cache_marked_cfields(self, site):
        marked_extensions = site.system_settings['configs.marked_extensions']
        if site.synamic.env['backend'] == 'file':  # TODO: fix it.
            content_service = site.get_service('contents')
            path_tree = self.get_path_tree(site)
            content_dir = site.synamic.system_settings['dirs.contents.contents']
            content_cdir = path_tree.create_dir_cpath(content_dir)
            # for content
            if content_cdir.exists():  # check content dir existence before proceeding
                file_cpaths = content_cdir.list_files()
                all_cfields = []

                # make the cfields
                for file_cpath in file_cpaths:
                    if file_cpath.extension.lower() in marked_extensions:
                        text = self.get_raw_text_data(site, file_cpath)
                        front_matter, body = content_splitter(file_cpath, text)
                        del body
                        try:
                            fields_syd = self.make_syd(front_matter)
                        except SynamicSydParseError as e:
                            raise SynamicErrors(
                                f'Synamic Syd parsing error during parsing front matter of file: '
                                f'{file_cpath.relative_path}\n'
                                f'<This error occurred during caching the cfileds of marked contents>',
                                e
                            )
                        cfields = content_service.build_cfields(fields_syd, file_cpath)
                        all_cfields.append(cfields)
                    else:
                        # No need to cache anything about static file.
                        pass

                # add the cfields to cache
                for cfields in all_cfields:
                    self.__cache.add_marked_cfields(site, cfields)

                # cache book toc
                book_tocs = []
                for cfields in all_cfields:
                    book_toc = cfields.book_toc
                    if book_toc is not None:
                        book_tocs.append(book_toc)
                # adding book toc to cache
                for book_toc in book_tocs:
                    self.__cache.add_book_toc(site, book_toc)

        else:
            raise NotImplemented
            # database backend is not implemented yet. AND there is nothing to do here for db, skip it when implemented

    def __cache_users(self, site):
        user_service = site.get_service('users')
        user_ids = user_service.get_user_ids()
        for user_id in user_ids:
            user = user_service.make_user(user_id)
            self.__cache.add_user(site, user)

    def __cache_markers(self, site):
        marker_service = site.get_service('markers')
        marker_ids = marker_service.get_marker_ids()
        for marker_id in marker_ids:
            marker = marker_service.make_marker(marker_id)
            self.__cache.add_marker(site, marker_id, marker)
            # self.__marker_by_id_cachemap[site.id][marker_id] = marker

    def __cache_pre_processed_contents(self, site):
        pre_processor_service = site.get_service('pre_processor')
        pre_processors = pre_processor_service.pre_processors
        for pre_processor in pre_processors:
            generated_contents = pre_processor.get_generated_contents()
            for generated_content in generated_contents:
                self.__cache.add_pre_processed_content(site, generated_content, generated_content.source_cpath)

        # TODO: keep sitemap here forever? IT will not make the system happy... it should be created before byproduct
        # ...or even after.
        # TODO: also, provide some mechanism to add content to sitemap dynamically.
        # TODO: so this is not the place for sitemap.
        sitemap_service = site.get_service('sitemap')
        sitemap_content = sitemap_service.make_sitemap(self.query_cfields(site, 'type != error :sortby updated_on desc'))
        self.__cache.add_pre_processed_content(site, sitemap_content)

    def __cache_menus(self, site):
        menu_service = site.get_service('menus')
        menu_names = menu_service.get_menu_names()
        for menu_name in menu_names:
            menu = menu_service.make_menu(menu_name)
            self.__cache.add_menu(site, menu_name, menu)

    def __cache_data(self, site):
        data_service = site.get_service('data')
        data_names = data_service.get_data_names()
        for data_name in data_names:
            data_instance = data_service.make_data(data_name)
            self.__cache.add_data(site, data_instance)

    def make_url_for_marked_content(self, site, file_cpath, path=None, slug=None, for_cdoctype=CDocType.TEXT_DOCUMENT):
        nourl_content_dirs_sw = self.get_site_settings(site)['configs.nourl_content_dirs_sw']
        if path is not None:
            #  discard everything and keep it. No processing needed.
            pass
        elif slug is not None:
            # calculate through file system.
            cpath_comps = file_cpath.path_comps
            basename = cpath_comps[-1]
            # TODO: what to do to basename? unused!
            cpath_comps = cpath_comps[1:-1]  # ignoring `contents` dir  TODO: make it more logical and dynamic
            # instead of hard coded.
            _ = []
            for ccomp in cpath_comps:
                if not ccomp.startswith(nourl_content_dirs_sw):
                    _.append(ccomp)

            cpath_comps = _
            cpath_comps.append(slug)
            path = cpath_comps
        else:
            # calculate through file system.
            cpath_comps = file_cpath.path_comps
            basename = cpath_comps[-1]
            cpath_comps = cpath_comps[1:-1]  # ignoring `contents` dir  TODO: make it more logical and dynamic
            # instead of hard coded.
            _ = []
            for ccomp in cpath_comps:
                if not ccomp.startswith(nourl_content_dirs_sw):
                    _.append(ccomp)

            cpath_comps = _
            cpath_comps.append(basename)
            path = cpath_comps

        url_path_comps = path
        curl = self.__synamic.router.make_url(
            site,
            url_path_comps,
            for_cdoctype=for_cdoctype
        )
        return curl

    def static_content_cpath_to_url(self, site, cpath, for_cdoctype):
        assert CDocType.is_binary(for_cdoctype, not_generated=True)
        # For STATIC Files
        contents_dir = site.system_settings['dirs.contents.contents']
        contents_cdir = site.path_tree.create_dir_cpath(contents_dir)
        basename_contents = contents_cdir.basename
        url_comps = cpath.path_comps
        url_comps = Sequence.lstrip(url_comps, (basename_contents,))
        curl = self.__synamic.router.make_url(
            site,
            url_comps,
            for_cdoctype=for_cdoctype
        )
        return curl

    # @loaded
    def get_cfields(self, site, path, default=None):
        path_tree = site.get_service('path_tree')
        path = path_tree.create_file_cpath(path)
        if site.synamic.env['backend'] == 'database':
            raise NotImplemented
        else:
            # file backend
            return self.__cache.get_marked_cfields_by_cpath(site, path, default=default)

    # @loaded
    def get_marked_content(self, site, path):
        # create content, meta, set meta with converters. Setting it from here will help caching.
        # TODO: other type of contents besides md contents.
        content_service = site.get_service('contents')
        if isinstance(path, str):
            path_tree = self.get_path_tree(site)
            contents_cdir = path_tree.create_dir_cpath(site.system_settings['dirs.contents.contents'])
            file_cpath = contents_cdir.join(path, is_file=True)
        else:
            file_cpath = path

        # check cache
        marked_content = self.__cache.get_marked_content_by_cpath(site, file_cpath, None)
        if marked_content is not None:
            return marked_content

        if not file_cpath.exists():
            return None
        else:
            cached_cfields = self.__cache.get_marked_cfields_by_cpath(site, file_cpath)
            assert cached_cfields is not None
            marked_content = content_service.build_md_content(file_cpath, cached_cfields)
            # cache it.
            self.__cache.add_marked_content(site, marked_content)
            return marked_content

    def get_marked_content_by_curl(self, site, curl):
        cpath = self.__cache.get_marked_cpath_by_curl(site, curl)
        if cpath is None:
            return None
        return self.get_marked_content(site, cpath)

    def get_pre_processed_content_by_curl(self, site, curl):
        return self.__cache.get_pre_processed_content_by_curl(site, curl, None)

    def get_marked_contents_by_cpaths(self, site, cpaths):
        contents = []
        for cpath in cpaths:
            contents.append(self.get_marked_content(site, cpath))
        return contents

    def get_binary_content(self, site, path):
        path_tree = site.get_service('path_tree')
        if not path_tree.exists(path):
            return None
        else:
            content_service = site.get_service('contents')
            return content_service.build_static_content(path)

    def get_static_file_cpaths(self, site):
        marked_extensions = site.system_settings['configs.marked_extensions']
        paths = []
        path_tree = site.get_service('path_tree')
        contents_dir = site.system_settings['dirs.contents.contents']
        contents_cdir = path_tree.create_dir_cpath(contents_dir)

        if contents_cdir.exists():
            for path in contents_cdir.list_files(checker=lambda cp: cp.extension not in marked_extensions):
                paths.append(path)
        return paths

    @staticmethod
    def empty_syd():
        return SydParser('').parse()

    def get_raw_text_data(self, site, path, encoding='utf-8') -> str:
        path_tree = self.get_path_tree(site)
        if not path_tree.is_type_cpath(path):
            path = path_tree.create_file_cpath(path)
        with path.open('r', encoding=encoding) as f:
            text = f.read()
        return text

    def get_syd(self, site, path):
        path_tree = site.path_tree
        if not path_tree.is_type_cpath(path):
            cpath = path_tree.create_file_cpath(path)
        else:
            cpath = path
        syd = self.__cache.get_syd(site, cpath, default=None)
        if syd is None and cpath.exists():
            try:
                syd = SydParser(self.get_raw_text_data(site, cpath)).parse()
            except (SynamicSydParseError, SynamicFSError) as e:
                raise SynamicErrors(
                    f'Synamic error during parsing syd file: '
                    f'{cpath.relative_path}',
                    e
                )
            self.__cache.add_syd(site, cpath, syd)
        return syd

    @staticmethod
    def make_syd(raw_data):
        syd = SydParser(raw_data).parse()
        return syd

    def get_model(self, site, model_name):
        """Nothing from the system model be overridden - this is opposite of system syd to user settings."""
        processed_model = self.__cache.get_model(site, model_name, None)
        if processed_model is None:
            types = site.get_service('types')

            model_dir = site.synamic.system_settings['dirs.metas.models']
            path_tree = site.get_service('path_tree')
            user_model_cpath = path_tree.create_file_cpath(model_dir, model_name + '.model')

            if user_model_cpath.exists():
                user_model = ModelParser.parse(model_name, self.get_raw_text_data(site, user_model_cpath))
            else:
                user_model = None

            system_model = site.default_data.get_model(model_name, None)
            if system_model is None and processed_model is None:
                return None
            elif system_model is not None and user_model is not None:
                new_model = user_model.new(system_model)
            elif system_model is None:
                new_model = user_model
            else:
                new_model = system_model
            # put converters inside of the cfields
            for key, field in new_model.items():
                converter = types.get_converter(field.converter_name)
                field.set_converter(converter)
            self.__cache.add_model(site, new_model)
            return new_model
        else:
            return processed_model

    def get_content_parts(self, site, content_path):
        text = self.get_raw_text_data(site, content_path)
        front_matter, body = content_splitter(content_path, text)
        front_matter_syd = self.make_syd(front_matter)  # Or take it from cache.
        return front_matter_syd, body

    def get_path_tree(self, site):
        return site.get_service('path_tree')

    def get_curl_by_filename(self, site, filename, relative_cpath=None, default=None):
        if relative_cpath is None:
            cpath = site.path_tree.create_file_cpath(filename, forgiving=True)
        else:
            if filename.startswith(('/', '\\')):
                cpath = site.path_tree.create_file_cpath(filename, forgiving=True)
            else:
                cpath = relative_cpath.parent_cpath.join(filename, is_file=True, forgiving=True)
        if not cpath.exists():
            return default
        content = self.__cache.get_marked_content_by_cpath(site, cpath, default=None)
        if content is None:
            pre_content = self.__cache.get_pre_processed_content_by_cpath(site, cpath, default=None)
        else:
            return content.curl
        if pre_content is None:
            static_content = site.get_service('contents').build_static_content(cpath)
            return static_content.curl

    def getc(self, site, url_str, relative_cpath=None):
        """
        For: curl:// url:// cfields:// content:// etc.
        """
        url_struct = parse_getc(url_str)
        result = None
        if url_struct.scheme in ('cfields', 'curl', 'geturl', 'url', 'content', 'cpath'):
            if not url_struct.keys:
                raise SynamicGetCParsingError(f'Malformated getc param: {url_str}')
            key = url_struct.keys[0]  # TODO: add support for more chained keys later.
            path = url_struct.path
            assert key is not None and path is not None

            # find the proper site
            site_id_sep = self.__synamic.system_settings['configs.site_id_sep']
            site_id = url_struct.site_id

            if site_id == site_id_sep:
                site = site
                c_res = self.__getc(site, key, path, None)
            elif site_id is not None:
                site_id_obj = self.__synamic.sites.make_id(site_id)
                try:
                    site = self.__synamic.sites.get_by_id(site_id_obj)
                except KeyError:
                    raise SynamicSiteNotFound(
                        f'Site with id {site_id} was not found when looking for it in getc() for {url_str} and relative'
                        f' path {relative_cpath}'
                    )
                c_res = self.__getc(site, key, path, None)
            else:
                c_res = None
                sites_up = self.sites_up(site)
                for specific_site in sites_up:
                    c_res = self.__getc(specific_site, key, path, None)
                    if c_res is not None:
                        site = specific_site
                        break

            if c_res is not None:
                if url_struct.scheme == 'cfields':
                    result = c_res['cfields']
                elif url_struct.scheme == 'curl':
                    result = c_res['cfields'].curl
                elif url_struct.scheme in ('url', 'geturl'):
                    result = c_res['cfields'].curl.url
                elif url_struct.scheme == 'cpath':
                    result = c_res['cfields'].cpath
                elif url_struct.scheme == 'content':
                    static_content = c_res.get('static_content', None)
                    scss_content = c_res.get('scss_content', None)  # CAUTION: sCss vs sAss
                    if static_content is not None:
                        result = static_content
                    elif scss_content is not None:
                        result = static_content
                    else:
                        # marked cfields
                        marked_cfields = c_res.get('marked_cfields', None)
                        if marked_cfields is not None:
                            result = self.get_marked_content(site, marked_cfields.cpath)
        else:
            if not url_struct.scheme:
                curl = self.get_curl_by_filename(site, url_str, relative_cpath=relative_cpath)
                if curl:
                    url_str = curl.url
            return url_str
        if result is not None:
            return result
        else:
            raise SynamicGetCError(f"Not found for getc ->  {url_str}")

    def __getc(self, specific_site, key, path, default=None):
        content_service = specific_site.get_service('contents')
        getc_key, getc_path = key, path

        contents_cdir = specific_site.path_tree.create_dir_cpath(
            specific_site.system_settings['dirs.contents.contents']
        )

        result_cfields = default
        if getc_key not in ('file', 'sass', 'id'):
            raise SynamicGetCError(
                f'GetC key error. Currently only file, sass, id keys are allowed, but key {getc_key} was provided'
            )
        if getc_key == 'file':
            file_cpath = contents_cdir.join(getc_path, is_file=True)
            if file_cpath.exists():
                marked_cfields = self.__cache.get_marked_cfields_by_cpath(
                    specific_site, file_cpath, None
                )
                if marked_cfields is not None:
                    result_cfields = {
                        'cfields': marked_cfields,
                        'marked_cfields': marked_cfields
                    }
                else:
                    # try STATIC
                    static_content = content_service.build_static_content(file_cpath)
                    result_cfields = {
                        'cfields': static_content.cfields,
                        'static_content': static_content
                    }
        elif getc_key == 'sass':
            # pre-processor stuff. Must be in pre processed content.
            scss_cpath = specific_site.get_service('pre_processor').get_processor('sass').make_cpath(getc_path)
            scss_content = self.__cache.get_pre_processed_content_by_cpath(specific_site, scss_cpath, None)
            if scss_content is not None:
                result_cfields = {
                    'cfields': scss_content.cfields,
                    'scss_content': scss_content  # CAREFUL: sCss vs sAss
                }
        elif getc_key == 'id':
            for cfields in self.__cache.get_all_marked_cfields(specific_site):
                if cfields.id == getc_path:
                    result_cfields = {
                        'cfields': cfields,
                        'marked_cfields': cfields
                    }
                    break
        return result_cfields

    def get_site_settings(self, site):
        ss = self.__site_settings.get(site.id, None)
        if ss is None:
            ss = site.get_service('site_settings').make_site_settings()
            self.__site_settings[site.id] = ss
        return ss

    def get_content_by_segments(self, site, path_segments, pagination_segments):
        """Method primarily for router.get()"""
        pass

    def get_marker(self, site, marker_id, default=None, error_out=False):
        marker = self.__cache.get_marker(site, marker_id, default=None)
        if marker is None and error_out:
            raise SynamicMarkerNotFound(f'Marker does not exist: {marker_id}')
        else:
            if marker is None:
                return default
            else:
                return marker

    def get_marker_by_slug(self, site, slug, default=None):
        markers = self.get_markers(site)
        for marker in markers:
            if marker.get('slug', default=None) == slug:
                return marker
        return default

    def get_markers_by_type(self, site, marker_type):
        allowed_marker_types = {'single', 'multiple', 'hierarchical'}
        if marker_type not in allowed_marker_types:
            raise SynamicMarkerNotFound(
                f'Marker of type {marker_type} is not allowed. Only allowed markers are'
                f' {", ".join(allowed_marker_types)}'
            )
        _ = []
        for marker in self.__cache.get_markers(site):
            if marker.type == marker_type:
                _.append(marker)
        return _

    def get_markers(self, site):
        return self.__cache.get_markers(site)

    def get_all_cached_marked_cfields(self, site):
        # TODO: logic for cached content metas
        # - when to use it when not (when not cached)
        return self.__cache.get_all_marked_cfields(site)

    def get_all_pre_processed_contents(self, site):
        return self.__cache.get_all_pre_processed_content(site)

    def get_marked_cpath_by_curl(self, site, curl, default=None):
        return self.__cache.get_marked_cpath_by_curl(site, curl, default=default)

    def get_menu(self, site, menu_name, default=None):
        return self.__cache.get_menu(site, menu_name, default=default)

    def get_menus(self, site):
        return self.__cache.get_menus(site)

    def get_data(self, site, data_name, default=None):
        data = None
        for site in self.sites_up(site):
            data = self.__cache.get_data(site, data_name, default=None)
            if data is not None:
                break

        if data is None:
            return default
        else:
            return data

    def get_book_tocs(self, site):
        return self.__cache.get_book_tocs(site)

    @staticmethod
    def __convert_section_value(section, content_model):
        # TODO: for converter that returns single value implement mechanism that will help use in !in for them.
        converted_value = content_model[section.key].converter(section.value)
        return SimpleQueryParser.QuerySection(key=section.key, comp_op=section.comp_op, value=converted_value)

    def __query_cfields_left_right(self, section, result_set, content_model):
        matched_result = set()
        for cfields in result_set:
            field_value = cfields.get(section.key, None)
            converter = content_model[section.key].converter
            if field_value is not None:
                if converter.compare(section.comp_op, field_value, section.value):
                    matched_result.add(cfields)
            else:
                if section.comp_op in ('!=', '!in'):
                    matched_result.add(cfields)
        return matched_result

    def __query_cfields_by_node(self, node, result_set, content_model):
        if isinstance(node, SimpleQueryParser.QuerySection):
            # only one section here.
            left_section = node
            right_section = None
            logic_op = None
        else:
            assert isinstance(node, QueryNode)
            logic_op = node.logic_op
            if isinstance(node.left, QueryNode):
                return self.__query_cfields_by_node(node.left, result_set, content_model)
            else:
                left_section = node.left
                right_section = None

            if isinstance(node.right, QueryNode):
                return self.__query_cfields_by_node(node.left, result_set, content_model)
            else:
                right_section = node.right

        left_section = self.__convert_section_value(left_section, content_model)
        left_result = self.__query_cfields_left_right(left_section, result_set, content_model)
        if right_section is not None:
            right_section = self.__convert_section_value(right_section, content_model)
            right_result = self.__query_cfields_left_right(right_section, result_set, content_model)
        else:
            right_result = None

        if logic_op == '|':
            left_result.update(right_result)
        elif logic_op == '&':
            left_result.intersection_update(right_result)
        else:
            # keep as is
            pass
        result_set.clear()
        result_set.update(left_result)
        return result_set

    def query_cfields(self, site, query_str):
        content_model = site.object_manager.get_model('content')
        node, sort = SimpleQueryParser(query_str).parse()
        all_cfields_s = self.__cache.get_all_marked_cfields(site)
        result = set(all_cfields_s)
        if node is not None:
            self.__query_cfields_by_node(node, result, content_model)
        if sort is not None:
            def sorting_key_func(f):
                value = f[sort.by_key]
                if value is None:
                    value = Nil
                return value
            result = sorted(
                result,
                key=sorting_key_func,
                reverse=True if sort.order == 'desc' else False
            )

        return tuple(result)

    def query_contents(self, site, query_str):
        cfields_s = self.query_cfields(site, query_str)
        _ = []
        for cfields in cfields_s:
            _.append(
                self.get_marked_content(site, cfields.cpath)
            )
        return tuple(_)

    def get_user(self, starting_site, user_id):
        user = None
        sites_up = self.sites_up(starting_site)
        for site in sites_up:
            user = self.__cache.get_user(site, user_id, None)
            if user is not None:
                break
        if user is None:
            raise SynamicUserNotFound(
                f'User with user id {user_id} not found in the {starting_site} site, neither in the '
                f'upper hierarchy.'
            )
        return user

    def get_users(self, site):
        return self.__cache.get_users(site)

    def sites_up(self, starting_site):
        """All sites up to the root including the current one"""
        sites_up = []
        site = starting_site
        while True:
            sites_up.append(site)
            if site.has_parent:
                site = site.parent
            else:
                break
        return sites_up

    class __ObjectManagerForSite:
        def __init__(self, site, object_manager):
            self.__site = site
            self.__object_manager = object_manager
            self.__is_loaded = False

            # real object manager methods.
            self.__cached_om_methods = {}

        def __getattr__(self, key):
            res = self.__cached_om_methods.get(key, None)
            if res is None:
                attr = getattr(self.__object_manager, key)
                if not isinstance(attr, types.MethodType):  # or key.startswith('_')
                    res = attr
                else:
                    om_method = attr

                    def call_om_method(*args, **kwargs):
                        return om_method(self.__site, *args, **kwargs)
                    self.__cached_om_methods[key] = call_om_method
                    res = call_om_method
            return res

        @property
        def is_loaded(self):
            return self.__is_loaded

        @loaded
        def reload(self):
            self.__is_loaded = False
            self.__object_manager.__reload_for__(self.site)
            self.__is_loaded = True

        @not_loaded
        def load(self):
            self.__object_manager.__load_for__(self.site)
            self.__is_loaded = True

        @property
        def site(self):
            return self.__site

    class __Cache:
        def __init__(self, synamic):
            self.__synamic = synamic

            # marker
            self.__marker_by_id_cachemap = defaultdict(dict)
            # syd cachemap
            self.__cpath_to_syd_cachemap = defaultdict(dict)

            # menus
            self.__menus_cachemap = defaultdict(dict)

            # models
            self.__models_cachemap = defaultdict(dict)

            # users
            self.__users_cachemap = defaultdict(dict)

            # data
            self.__data = defaultdict(dict)

            # >>>>>>>>
            # CONTENTS
            self.__pre_processed_cachemap = defaultdict(dict)  # curl to content
            self.__marked_contents_cachemap = defaultdict(dict)  # curl to content: default limit 100 per site. < limit not yet implemented
            self.__marked_cfields_cachemap = defaultdict(dict)  # curl to content
            self.__book_tocs_cachemap = defaultdict(dict)

            self.__cpath_to_pre_processed_contents = defaultdict(dict)
            self.__cpath_to_marked_content = defaultdict(dict)
            self.__cpath_to_marked_cfields = defaultdict(dict)
            # <<<<<<<<

        def add_pre_processed_content(self, site, pre_processed_content, cpath=None):
            """Pre processed content is one kind of generated content, sor we cannot rely on source-cpath and thus cpath
            parameter is needed explicitly"""
            self.__pre_processed_cachemap[site.id][pre_processed_content.curl] = pre_processed_content
            if cpath is not None:
                self.__cpath_to_pre_processed_contents[site.id][cpath] = pre_processed_content

        def get_pre_processed_content_by_curl(self, site, curl, default=None):
            res = self.__pre_processed_cachemap[site.id].get(curl, default)
            return res

        def get_pre_processed_content_by_cpath(self, site, cpath, default=None):
            res = self.__cpath_to_pre_processed_contents[site.id].get(cpath, default)
            return res

        def get_all_pre_processed_content(self, site):
            return tuple(self.__pre_processed_cachemap[site.id].values())

        def add_marked_content(self, site, marked_content):
            # TODO: set limit to 100 or so
            self.__marked_contents_cachemap[site.id][marked_content.curl] = marked_content
            self.__cpath_to_marked_content[site.id][marked_content.cpath] = marked_content

        def get_marked_content_by_curl(self, site, curl, default=None):
            return self.__marked_contents_cachemap[site.id].get(curl, default)

        def get_marked_content_by_cpath(self, site, cpath, default=None):
            return self.__cpath_to_marked_content[site.id].get(cpath, default)

        def add_marked_cfields(self, site, cfields):
            self.__marked_cfields_cachemap[site.id][cfields.curl] = cfields
            self.__cpath_to_marked_cfields[site.id][cfields.cpath] = cfields

        def get_marked_cfields_by_curl(self, site, curl, default=None):
            return self.__marked_cfields_cachemap[site.id].get(curl, default)

        def get_marked_cfields_by_cpath(self, site, cpath, default=None):
            return self.__cpath_to_marked_cfields[site.id].get(cpath, default)

        def get_marked_cpath_by_curl(self, site, curl, default=None):
            cfs = self.get_marked_cfields_by_curl(site, curl, None)
            if cfs is None:
                return default
            else:
                return cfs.cpath

        def get_all_marked_cfields(self, site):
            return tuple(self.__marked_cfields_cachemap[site.id].values())

        def add_marker(self, site, marker_id, marker):
            self.__marker_by_id_cachemap[site.id][marker_id] = marker

        def get_marker(self, site, marker_id, default=None):
            return self.__marker_by_id_cachemap[site.id].get(marker_id, default)

        def get_markers(self, site):
            return tuple(self.__marker_by_id_cachemap[site.id].values())

        def add_syd(self, site, cpath, syd):
            self.__cpath_to_syd_cachemap[site.id][cpath] = syd

        def get_syd(self, site, cpath, default=None):
            return self.__cpath_to_syd_cachemap[site.id].get(cpath, default)

        def add_menu(self, site, menu_name, menu):
            self.__menus_cachemap[site.id][menu_name] = menu

        def get_menu(self, site, menu_name, default=None):
            menu = self.__menus_cachemap[site.id].get(menu_name, None)
            if menu is None:
                return default
            return menu

        def get_menus(self, site):
            return tuple(self.__menus_cachemap[site.id].values())

        def add_model(self, site, model):
            self.__models_cachemap[site.id][model.model_name] = model
            return model

        def get_model(self, site, model_name, default=None):
            return self.__models_cachemap[site.id].get(model_name, default)

        def add_user(self, site, user):
            self.__users_cachemap[site.id][user.id] = user

        def get_user(self, site, user_id, default=None):
            return self.__users_cachemap[site.id].get(user_id, default)

        def get_users(self, site):
            return tuple(self.__users_cachemap[site.id].values())

        def add_data(self, site, data):
            self.__data[site.id][data.get_data_name()] = data

        def get_data(self, site, data_name, default=None):
            return self.__data[site.id].get(data_name, default)

        def add_book_toc(self, site, book_toc):
            book_content_cpath = book_toc.book_cpath
            self.__book_tocs_cachemap[site.id][book_content_cpath] = book_toc

        def get_book_tocs(self, site):
            return tuple(self.__book_tocs_cachemap[site.id].values())

        def clear_content_cache(self, site):
            self.__pre_processed_cachemap[site.id].clear()
            self.__marked_contents_cachemap[site.id].clear()
            self.__marked_cfields_cachemap[site.id].clear()
            self.__cpath_to_pre_processed_contents[site.id].clear()
            self.__cpath_to_marked_content[site.id].clear()
            self.__cpath_to_marked_cfields[site.id].clear()
            self.__book_tocs_cachemap[site.id].clear()

        def clear_marker_cache(self, site):
            self.__marker_by_id_cachemap[site.id].clear()

        def clear_syd_cache(self, site):
            self.__cpath_to_syd_cachemap[site.id].clear()

        def clear_menus_cache(self, site):
            self.__menus_cachemap[site.id].clear()

        def clear_model(self, site):
            self.__models_cachemap[site.id].clear()

        def clear_users(self, site):
            self.__users_cachemap[site.id].clear()

        def clear_data(self, site):
            self.__data[site.id].clear()

        def clear_cache(self, site):
            """Clear all"""
            self.clear_content_cache(site)
            self.clear_marker_cache(site)
            self.clear_syd_cache(site)
            self.clear_menus_cache(site)
            self.clear_model(site)
            self.clear_users(site)
            self.clear_data(site)

    def build(self, site):
        output_dir = self.__synamic.system_settings['dirs.outputs.outputs']
        output_cdir = self.__synamic.path_tree.create_dir_cpath(output_dir)

        c_iter = CIter(site)
        for content in c_iter:
            curl = content.curl
            print(f'Writing {curl.url}')
            with content.get_stream() as fr:
                c_out_dir, fn = curl.to_dirfn_pair_w_site
                c_out_cdir = output_cdir.join(c_out_dir, is_file=False)
                if not c_out_cdir.exists():
                    c_out_cdir.makedirs()
                c_out_cfile = c_out_cdir.join(fn, is_file=True)
                with c_out_cfile.open('wb') as fw:
                    data = fr.read(1024)
                    while data:
                        fw.write(data)
                        data = fr.read(1024)
        return True

    def init_site(self, site=None):
        assert site is None, 'Must not be called on a site instance. It should be called when synamic is not init\'ed'
        assert not self.__synamic.is_loaded
        root_cpath = self.__synamic.path_tree.create_dir_cpath('/')
        init_manager = InitManager(self.__synamic)
        return init_manager.init_site(root_cpath)


class CIter:
    def __init__(self, site):
        self.__site = site

    def __make_clist(self):
        content_service = self.__site.get_service('contents')
        # marked
        clist = []

        all_cfields = self.__site.object_manager.get_all_cached_marked_cfields()
        all_pre_content = self.__site.object_manager.get_all_pre_processed_contents()
        all_static_cpaths = self.__site.object_manager.get_static_file_cpaths()
        all_users = self.__site.object_manager.get_users()
        all_marks = []

        for marker in self.__site.object_manager.get_markers():
            if not marker.is_public:
                continue
            marks = marker.marks
            for mark in marks:
                all_marks.append(mark.content)

        clist.extend(all_cfields)
        clist.extend(all_pre_content)
        clist.extend(all_static_cpaths)
        clist.extend(all_users)
        clist.extend(all_marks)

        # pagination pages
        pagination_pages = []
        for cfields in all_cfields:
            root_pagination = cfields.pagination
            if content_service.is_type_pagination_page(root_pagination):
                for page_no in range(1, root_pagination.total_pagination):
                    sub_page = root_pagination.get_sub_page(page_no)
                    assert content_service.is_type_pagination_page(sub_page), f'Type: {type(sub_page)}'
                    pagination_pages.append(sub_page)
        clist.extend(pagination_pages)
        return clist

    def __iter__(self):
        return self.__CListIterator(self.__make_clist(), self.__site)

    class __CListIterator:
        def __init__(self, clist, site):
            self.__clist = clist
            self.__idx = 0
            self.__site = site

            self.__content_service = self.__site.get_service('contents')
            self.__markers_service = self.__site.get_service('markers')
            self.__users_service = self.__site.get_service('users')
            self.__path_tree = self.__site.path_tree

        def __next__(self):
            if self.__idx >= len(self.__clist):
                del self.__clist
                del self.__site
                del self.__content_service
                del self.__markers_service
                del self.__users_service
                del self.__path_tree
                raise StopIteration

            elem = self.__clist[self.__idx]
            self.__idx += 1

            # marked content
            if self.__content_service.is_type_cfields(elem):
                cfields = elem
                content = self.__site.object_manager.get_marked_content(cfields.cpath)

            elif self.__content_service.is_type_generated_content(elem):
                content = elem

            # static content
            elif self.__path_tree.is_type_cpath(elem):
                cpath = elem
                content = self.__content_service.build_static_content(cpath)

            # user
            elif self.__users_service.is_type_user(elem):
                user = elem
                content = user.content

            # mark
            elif self.__markers_service.is_type_mark(elem):
                mark = elem
                content = mark.content

            # pagination pages
            elif self.__content_service.is_type_pagination_page(elem):
                pagination_page = elem
                content = pagination_page.host_content

            else:
                raise Exception(f'Something impossible happened or you introduced a bug.')

            return content



__all__ = ['Nil']


class _Nil:
    __instance_count = 0

    def __init__(self):
        assert self.__instance_count == 0, f'Only one instance of Nil can exist in the whole system. Do not try to ' \
                                           f'instantiate by yourself. Use the one synamic.Nil'

        self.__instance_count += 1

    def __eq__(self, other):
        return id(other) == id(self)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __gt__(self, other):
        return False

    def __lt__(self, other):
        return False

    def __ge__(self, other):
        return self.__eq__(other)

    def __le__(self, other):
        return self.__eq__(other)

    def __contains__(self, item):
        return False

    def __hash__(self):
        return id(self)

    def __bool__(self):
        return False

    def __str__(self):
        return ''

    def __repr__(self):
        return ''

    def __getattr__(self, item):
        return self


Nil = _Nil()



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""


from .base_shell import BaseShell
import sys


class CommandProcessor(BaseShell):
    intro_text = 'Welcome to the Synamic shell.   Type help or ? to list commands.\n'
    prompt_text = '(synamic)$ '

    def __init__(self, synamic_class, start_dir, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__global_for_py = {}
        self.__synamic_object = None
        self.__synamic_class = synamic_class
        self.__root_site_root = start_dir

    def get_or_create_synamic(self):
        if self.__synamic_object is None:
            self.__synamic_object = self.__synamic_class(self.__root_site_root)
        return self.__synamic_object

    def on_cwd(self):
        self.pprint("Site Root: %s" % self.__root_site_root)

    def on_urls(self):
        raise NotImplemented

    def on_settings(self, arg):
        if not arg:
            self.pprint(self.__synamic_object.settings.values())
        else:
            val = self.__synamic_object.settings.get(arg, 'null')
            self.print(val)

    def on_init(self, *args):
        'Initialize a synamic project'
        s = self.get_or_create_synamic()
        ret = s.object_manager.init_site()
        if ret is True:
            self.print('Site successfully init\'ed')

    def on_load(self, *args):
        'Reload the current synamic project'
        s = self.get_or_create_synamic()
        if not s.is_loaded:
            s.load()

    def on_build(self, *args):
        'Build Synamic project that will result in static site'
        o = self.get_or_create_synamic()
        if not o.is_loaded:
            o.load()
        return o.sites.build()

    def on_reset(self):
        self.__synamic_object = None
        self.print(f'Synamic object set to none so that new instance start from the beginning')

    def on_serve(self, *args):
        '''Serve the current synamic project in localhost'''
        from synamic.entry_points.aio_server import aio_server

        SITE_ROOT = self.__root_site_root
        host = 'localhost'
        port = '8087'
        return aio_server.serve(SITE_ROOT, host, port)

    def on_upload(self, *args):
        'Deploy the build'
        if len(args) < 1:
            self.print_error(f'No upload parameter provided')
            exit(1)
        uploader_name = args[0].strip()
        synamic = self.get_or_create_synamic()
        if not synamic.is_loaded:
            synamic.load()
        uploader = synamic.upload_manager.get_uploader(uploader_name, None)
        if uploader is None:
            self.print_error(f'No uploader found with name {uploader_name}')
            exit(1)
        else:
            upload_res = uploader.upload()
            if upload_res:
                exit(0)
            else:
                exit(1)


def start_shell(synamic_class, start_dir):
    CommandProcessor(synamic_class, start_dir).run_command(sys.argv[1:])



import abc


class CFieldsContract(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def get(self, key, default=None):
        pass

    @abc.abstractmethod
    def set(self, key, value):
        pass

    @abc.abstractmethod
    def keys(self):
        pass

    @property
    @abc.abstractmethod
    def raw(self):
        pass

    @property
    @abc.abstractmethod
    def cdoctype(self):
        """
        Instance of document __cdoc_types enum
        """
        pass

    @property
    @abc.abstractmethod
    def cpath(self):
        pass

    @property
    @abc.abstractmethod
    def curl(self):
        pass

    @property
    @abc.abstractmethod
    def cmodel(self):
        pass

    @property
    @abc.abstractmethod
    def mimetype(self):
        pass



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""
import io
from synamic.core.contracts.content import ContentContract, CDocType


class GeneratedContent(ContentContract):
    def __init__(self, site, synthetic_cfields, file_content, source_cpath=None, render_callable=None):

        self.__site = site
        self.__synthetic_cfields = synthetic_cfields
        self.__file_content = file_content
        self.__source_cpath = source_cpath
        self.__render_callable = render_callable  # render callable will accept 2 params: site, this content object

        # validation
        assert CDocType.is_generated(self.__synthetic_cfields.cdoctype)
        assert type(self.__file_content) in (type(None), bytes, bytearray, str)
        assert synthetic_cfields.cpath is None
        assert self.__file_content is not None or source_cpath is not None

    @property
    def site(self):
        return self.__site

    @property
    def cfields(self):
        return self.__synthetic_cfields

    @property
    def __get_content_as_str(self):
        if self.__file_content is not None:
            content = self.__file_content
            if isinstance(content, (bytearray, bytes)):
                content = content.decode('utf-8')
        else:
            with self.__source_cpath.open('r') as f:
                content = f.read()
        return content

    def get_stream(self):
        if callable(self.__render_callable):
            # so this content is renderable.
            text_content = self.__render_callable(self.__site, self)
            stream = io.BytesIO(text_content.encode('utf-8'))
        else:
            if self.__file_content is not None:
                if isinstance(self.__file_content, (bytes, bytearray)):
                    stream = io.BytesIO(self.__file_content)
                else:
                    stream = io.BytesIO(self.__file_content.encode('utf-8'))
            else:
                stream = self.__source_cpath.open('rb')
        return stream

    @property
    def body(self):
        if self.__file_content is not None:
            body = self.__file_content
        else:
            file_mode = 'r'
            if CDocType.is_binary(self.cfields.cdoctype):
                file_mode = 'rb'
            with self.__source_cpath.open(file_mode) as f:
                body = f.read()
        return body

    @property
    def body_as_bytes(self):
        body = self.body
        if not isinstance(body, (bytes, bytearray)):
            body = body.encode('utf-8')
        return body

    @property
    def body_as_string(self):
        body = self.body
        if not isinstance(body, str):
            body = body.decode('utf-8')
        return body

    def __getitem__(self, key):
        return self.__synthetic_cfields[key]

    def __getattr__(self, key):
        return getattr(self.__synthetic_cfields, key)

    def __str__(self):
        return "Generated content: <%s>\n" % str(self.cfields.curl) + '...'

    def __repr__(self):
        return str(self)

    @property
    def source_cpath(self):
        return self.__source_cpath




import operator
import sly
from sly import Lexer, Parser
from synamic.exceptions import SynamicQueryParsingError
from synamic.core.parsing_systems.curlybrace_parser import SydParser  #covert_one_value
from collections import namedtuple


def generate_error_message(text, text_rest):
    err_before_txt = '_' * (len(text) - len(text_rest))
    err_after_txt = '^' * len(text_rest)
    err_txt = '\n' + text + '\n' + err_before_txt + err_after_txt
    return err_txt


class SimpleQueryParser:
    COMPARE_OPERATORS = tuple([
        'contains',
        '!contains',
        'in',
        '!in',
        '==',
        '!=',
        '>=',
        '<=',
        '>',
        '<',
    ])

    COMPARE_OPERATORS_SET = frozenset(COMPARE_OPERATORS)

    COMPARE_OPERATOR_2_PY_OPERATOR_FUN = {
        'contains': operator.contains,
        '!contains': lambda a, b: not operator.contains(a, b),
        'in': lambda a, b: operator.contains(b, a),
        '!in': lambda a, b: not operator.contains(b, a),
        '==': operator.eq,
        '!=': operator.ne,
        '>=': operator.ge,
        '<=': operator.le,
        '>': operator.gt,
        '<': operator.lt
    }

    Query = namedtuple('Query', ('node', 'sort'))
    QuerySection = namedtuple('QuerySection', ('key', 'comp_op', 'value'))
    QuerySortBy = namedtuple('QuerySortBy', ('by_key', 'order'))

    def __init__(self, txt):
        self.__txt = txt

    def parse(self):
        """
        title == something | type in go, yes & age > 6
        """
        text = self.__txt.strip()
        lexer = QueryLexer()
        parser = QueryParser(text)
        try:
            if text == '':
                res = self.Query(None, None)
            else:
                res = parser.parse(lexer.tokenize(text))

            if len(res) == 1:
                if isinstance(res[0], self.QuerySortBy):
                    query = self.Query(node=None, sort=res[0])
                else:
                    query = self.Query(node=res[0], sort=None)
            else:
                assert len(res) > 1
                query = self.Query(node=res[0], sort=res[1])
        except sly.lex.LexError as e:
            err_txt = generate_error_message(text, e.text)
            raise SynamicQueryParsingError(
                f'Lexical error at index: {e.error_index}\nDetails:{err_txt}'
            )
        else:
            return query


class QueryNode:
    def __init__(self, logic_op, left, right):
        self.__logic_op = logic_op
        self.__left = left
        self.__right = right

    @property
    def left(self):
        return self.__left

    @property
    def right(self):
        return self.__right

    @property
    def logic_op(self):
        return self.__logic_op

    def __str__(self):
        return """
        QueryNode( %s %s %s )
        """ % (str(self.left), str(self.logic_op), str(self.right))

    def __repr__(self):
        return self.__str__()


def test(query):
    tokens = SimpleQueryParser(query).parse()
    for token in tokens:
        print(token)


if __name__ == 'l__main__':
    test('up sort asc | x == 1&time > 11:24 PM | date == 2013-2-1 | dt < 1023-12-12     11:33:43 am | a > b & c in d | d in ~hh        & m contains tag 1, tag 2 & n !in go sfsdfsdf')


class QueryValueLexer(Lexer):
    tokens = {'VALUE', }

    @_(r'[^&|:]+')
    def VALUE(self, t):
        self.begin(QueryLexer)
        t.value = t.value.strip()
        return t


class QueryLexer(Lexer):
    tokens = {'KEY', 'COMP_OP', 'AND', 'OR', 'SORT_BY'}
    ignore_ws = r'\s'
    SORT_BY = r':sortby\s+'

    @_(r'>|<|==|!=|>=|<=|\s+in|!in|\s+contains|!contains')
    def COMP_OP(self, t):
        self.begin(QueryValueLexer)
        t.value = t.value.strip()
        return t

    KEY = r'[a-zA-Z_][a-zA-Z0-9_]*(\.[a-zA-Z_][a-zA-Z0-9_]*)*'

    KEY['contains'] = 'COMP_OP'
    KEY['in'] = 'COMP_OP'

    @_(r'&')
    def AND(self, t):
        return t

    @_(r'\|')
    def OR(self, t):
        return t


class QueryParser(Parser):
    def __init__(self, text):
        self.__text = text

    # Get the token list from the lexer (required)
    tokens = {*QueryLexer.tokens, *QueryValueLexer.tokens}
    precedence = [
        ('left', 'OR'),
        ('left', 'AND'),
    ]

    # Grammar rules and actions
    @_('expr',
       'sort',
       'expr sort')
    def query(self, p):
        if len(p) == 1:
            return p[0],
        else:
            return p[0], p[1]

    @_('SORT_BY KEY',
       'SORT_BY KEY KEY')
    def sort(self, p):
        if len(p) == 2:
            order = 'asc'
        else:
            if p[2].startswith('a'):
                order = 'asc'
            else:
                order = 'desc'
        return SimpleQueryParser.QuerySortBy(by_key=p[1], order=order)

    @_('KEY COMP_OP VALUE')
    def expr(self, p):
        converted_value = SydParser.covert_one_value(p[2])
        return SimpleQueryParser.QuerySection(key=p[0], comp_op=p[1], value=converted_value)

    @_('expr OR expr',
       'expr AND expr')
    def expr(self, p):
        left_section, right_section = p[0], p[2]
        return QueryNode(p[1], left_section, right_section)

    def error(self, p):
        text = self.__text
        if not p:
            text_rest = ''
            err_txt = generate_error_message(text, text_rest)
            err = SynamicQueryParsingError(
                f'End of query string before tokens could be parsed sensibly.\nDetails:{err_txt}'
            )
        else:
            text_rest = self.__text[p.index:]
            err_txt = generate_error_message(text, text_rest)
            err = SynamicQueryParsingError(
                f'Parsing error at token {p.type}.\nDetails:{err_txt}'
            )
        raise err


if __name__ == '__main__':
    while True:
        try:
            text = input('parse > ')
        except EOFError:
            raise
        if text:
            try:
                res = SimpleQueryParser(text).parse()
            except SynamicQueryParsingError as e:
                print(e.message)
            else:
                print(res)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

import os
import jinja2
from synamic.core.services.template.template_tags import GetCExtension, ResizeImageExtension
from synamic.core.standalones.functions.decorators import loaded, not_loaded
from synamic.core.standalones.functions.parent_config_splitter import parent_config_str_splitter
from .loaders import SynamicJinjaFileSystemLoader
from synamic.exceptions import SynamicTemplateError


class SynamicTemplateService:
    def __init__(self, site):
        self.__is_loaded = False
        self.__site = site
        self.__template_env = None

    @not_loaded
    def load(self):
        self.__template_env = jinja2.Environment(
            loader=SynamicJinjaFileSystemLoader(self.__site),
            autoescape=jinja2.select_autoescape(['html', 'xml']),
            extensions=[GetCExtension, ResizeImageExtension]
        )
        # setting config object to global of environment
        self.__template_env.site_object = self.__site
        # self.__template_env.globals['site'] = self.__site

        self.__is_loaded = True

    @property
    def is_loaded(self):
        return self.__is_loaded

    def exists(self, template_name):
        return True if os.path.exists(os.path.join(self.__site.template_dir, template_name)) else False

    @loaded
    def render(self, template_name, context=None, **kwargs):
        context = {} if context is None else context
        context.update(kwargs)

        # is_from_parent, template_name = parent_config_str_splitter(template_name)

        try:
            template = self.__template_env.get_template(template_name)
            result = template.render(context)
        except jinja2.exceptions.TemplateError as e:
            raise SynamicTemplateError(e)

        return result



import os
from synamic.core.parsing_systems.curlybrace_parser import SydParser
from synamic.core.parsing_systems.model_parser import ModelParser
_BASE_DIR = os.path.dirname(os.path.abspath(__file__))


def get_file_contents(fn, default=None):
    full_fn = os.path.join(_BASE_DIR, fn)
    if not os.path.exists(full_fn):
        return default
    with open(full_fn, encoding='utf-8') as f:
        text = f.read()
    return text


class DefaultDataManager:
    def __init__(self):
        self.__loaded_syds = {}
        self.__loaded_models = {}
        self.__system_settings = None

    def get_syd(self, name, default=None):
        if name in self.__loaded_syds:
            sydC = self.__loaded_syds[name]
        else:
            text = get_file_contents(name + '.syd', None)
            if text is None:
                return default
            sydC = SydParser(text).parse()
            self.__loaded_syds[name] = sydC
        return sydC

    def get_system_settings(self):
        if self.__system_settings is None:
            configs_syd = self.get_syd('configs')
            dirs_syd = self.get_syd('dirs')
            settings_syd = self.get_syd('settings')
            system_settings_syd = configs_syd.new(dirs_syd, settings_syd)
            self.__system_settings = system_settings_syd
        return self.__system_settings

    def get_model(self, name, default=None):
        if name in self.__loaded_models:
            model = self.__loaded_models[name]
        else:
            text = get_file_contents(name + '.model', None)
            if text is None:
                return default
            model = ModelParser.parse(name, text)
            self.__loaded_models[name] = model
        return model



import io
from synamic.core.contracts import ContentContract, CDocType


class PaginatedContent(ContentContract):
    def __init__(self, site, origin_cfields, paginated_cfields):
        self.__site = site
        self.__origin_cfields = origin_cfields
        self.__cfields = paginated_cfields
        self.__model = paginated_cfields.cmodel

        # validation
        assert CDocType.is_text(self.__cfields.cdoctype), f'Wrong Doctype {self.__cfields.cdoctype}'

    @property
    def site(self):
        return self.__site

    def __rendered(self):
        template_name = self.__cfields.get('template', 'default.html')
        templates = self.__site.get_service('templates')
        res = templates.render(template_name, context={
            'site': self.__site,
            'content': self
        })
        return res

    def get_stream(self):
        res = self.__rendered()
        f = io.BytesIO(res.encode('utf-8'))
        return f

    @property
    def body(self):
        content = self.__site.object_manager.get_marked_content(self.__origin_cfields.cpath)
        assert content is not None
        return content.body

    @property
    def body_as_string(self):
        return self.body

    @property
    def body_as_bytes(self):
        return self.body.encode('utf-8')

    @property
    def cfields(self):
        return self.__cfields

    def __getitem__(self, key):
        return self.__cfields[key]

    def __getattr__(self, key):
        return getattr(self.__cfields, key)

    def __str__(self):
        return "Paginated content: <%s>\n" % self['title'] + '...'

    def __repr__(self):
        return str(self)


class PaginationPage:
    def __init__(self, site, total_pagination, cfields_s, position, per_page):
        # assert len(cfields_s) != 0
        self.__site = site
        self.__total_pagination = total_pagination
        self.__cfields_s = cfields_s
        self.__position = position
        self.__per_page = per_page

        assert position < total_pagination
        # settable once.
        self.__host_content = None
        self.__next_content = None
        self.__prev_content = None

        # only available to root page/pagination
        self.__sub_pages = None

    @property
    def total_pagination(self):
        return self.__total_pagination

    @property
    def cfields_s(self):
        return self.__cfields_s

    @property
    def contents(self):
        if self.__cfields_s:
            contents = self.__site.object_manager.get_marked_contents_by_cpaths(
                cfields.cpath for cfields in self.__cfields_s
            )
            assert contents
        else:
            contents = tuple()

        return contents

    @property
    def position(self):
        return self.__position

    @property
    def per_page(self):
        return self.__per_page

    @property
    def host_content(self):
        return self.__host_content

    @host_content.setter
    def host_content(self, hos_con):
        assert self.__host_content is None
        self.__host_content = hos_con

    # Synthetic ones

    @property
    def is_start(self):
        if self.__position == 0:
            return True
        return False

    @property
    def is_end(self):
        if self.__position == self.__total_pagination - 1:
            return True
        else:
            return False

    @property
    def is_only(self):
        if self.__total_pagination == 1:
            return True
        return False

    @property
    def has_next(self):
        if self.__position < self.__total_pagination - 1:
            return True
        return False

    @property
    def next_content(self):
        return self.__next_content

    @next_content.setter
    def next_content(self, np):
        assert self.__next_content is None
        self.__next_content = np

    @property
    def has_previous(self):
        if self.__position > 0:
            return True
        return False

    @property
    def previous_content(self):
        return self.__prev_content
    prev_content = previous_content

    @previous_content.setter
    def previous_content(self, pp):
        assert self.__prev_content is None
        self.__prev_content = pp

    @property
    def sub_pages(self):
        assert self.is_start, "Cannot use this method for non root pages"

        sub_pages = self.__sub_pages
        if sub_pages is None:
            assert self.__per_page == 0, "Sub pages was not set after creation or per_page was set to wrong value"
            return tuple()
        else:
            return sub_pages

    @sub_pages.setter
    def sub_pages(self, sub_pages):
        assert self.is_start, "Cannot use this method for non root pages"

        assert self.__sub_pages is None
        assert isinstance(sub_pages, (list, tuple))
        self.__sub_pages = sub_pages

    def get_sub_page(self, idx, default=None):
        idx = idx - 1
        assert idx >= 0, "Root/start pagination is considered 0 and this method is only available on root"
        sub_pages = self.sub_pages
        if idx >= len(sub_pages):
            return default
        else:
            return sub_pages[idx]

    @property
    def is_empty(self):
        return self.__bool__()

    def __bool__(self):
        return self.__per_page > 0

    def __str__(self):
        return "Pagination Page: %d" % self.__position

    def __repr__(self):
        return repr(self.__str__())

    @classmethod
    def paginate_cfields(cls, site, origin_content, queried_cfields_s, per_page):
        site_settings = site.settings
        url_partition_comp = site_settings['url_partition_comp']
        pagination_url_comp = site_settings['pagination_url_comp']

        paginated_cfields_s_divisions = []

        if queried_cfields_s:
            quotient, remainder = divmod(len(queried_cfields_s), per_page)
            divisions = quotient
            if remainder > 0:
                divisions += 1

            for division_idx_i in range(divisions):
                division_cfields_s = []
                for idx_in_division_j in range(per_page):
                    idx = (division_idx_i * per_page) + idx_in_division_j  # (row * NUMCOLS) + column        #(i * divisions) + j
                    if idx >= len(queried_cfields_s):
                        break

                    # creating paginated content
                    original_cfields = queried_cfields_s[idx]
                    division_cfields_s.append(original_cfields)
                paginated_cfields_s_divisions.append(tuple(division_cfields_s))

        paginations = []
        paginated_contents = []
        if paginated_cfields_s_divisions:
            prev_page = None
            for division_idx_i, division_cfields_s in enumerate(paginated_cfields_s_divisions):
                pagination = PaginationPage(
                    site,
                    len(paginated_cfields_s_divisions),
                    division_cfields_s,
                    division_idx_i,
                    per_page
                )
                paginations.append(pagination)

                if division_idx_i == 0:
                    # setting pagination to the host content
                    origin_content.cfields.set('pagination', pagination)

                    pagination.host_content = origin_content
                    prev_page = origin_content

                else:
                    # creating paginated content
                    cdoctype = CDocType.GENERATED_HTML_DOCUMENT
                    curl = origin_content.curl.join(
                        "/%s/%s/%d/" % (url_partition_comp, pagination_url_comp, division_idx_i),
                        for_cdoctype=cdoctype)
                    paginated_cfields = origin_content.cfields.as_generated(
                        curl, cdoctype=cdoctype
                    )
                    paginated_cfields.set(
                        'title',
                        origin_content.cfields.get('title') + " - %s %d" % (pagination_url_comp.title(), division_idx_i)
                    )
                    paginated_content = PaginatedContent(
                        site,
                        origin_content.cfields,
                        paginated_cfields
                    )

                    # setting pagination to an aux content
                    paginated_content.cfields.set('pagination', pagination)
                    pagination.host_content = paginated_content

                    pagination.previous_content = prev_page
                    # TODO: content wrapper for prev/next page
                    prev_page.pagination.next_content = paginated_content
                    prev_page = paginated_content

                    paginated_contents.append(paginated_content)

        if not paginations:
            pagination = PaginationPage(
                site,
                1,
                tuple(),
                0,
                0
            )
            origin_content.cfields.set('pagination', pagination)
            paginations.append(pagination)
        paginations[0].sub_pages = paginations[1:]
        return tuple(paginations), tuple(paginated_contents)



"""
    author: "Md. Sabuj Sarker"
    copyright: "Copyright 2017-2018, The Synamic Project"
    credits: ["Md. Sabuj Sarker"]
    license: "MIT"
    maintainer: "Md. Sabuj Sarker"
    email: "md.sabuj.sarker@gmail.com"
    status: "Development"
"""

from synamic.core.standalones.functions.decorators import not_loaded, loaded
from .builtin_processors import _builtin_processor_classes
from synamic.exceptions import SynamicPreProcessorNotFound


class PreProcessorService:
    def __init__(self, site):
        self.__site = site

        self.__name_to_processor = {}

        self.__is_loaded = False

    @property
    def is_loaded(self):
        return self.__is_loaded

    @not_loaded
    def load(self):
        # load builtin processor
        path_tree = self.__site.get_service('path_tree')
        preprocess_cdir = path_tree.create_dir_cpath(self.__site.system_settings['dirs.contents.pre_process'])
        if preprocess_cdir.exists():
            cdirs = preprocess_cdir.list_dirs(depth=1)
            for cdir in cdirs:
                processor_name = cdir.basename
                if processor_name in _builtin_processor_classes:
                    self.add_processor(processor_name, cdir, _builtin_processor_classes[processor_name])
        # Add builtin pre-processor even if the dir does not exist for it
        for processor_name, processor_class in _builtin_processor_classes.items():
            if self.get_processor(processor_name, default=None, error_out=False) is None:
                cdir = preprocess_cdir.join(processor_name, is_file=False)
                self.add_processor(processor_name, cdir, processor_class)

        for processor in self.__name_to_processor.values():
            processor.load()
        self.__is_loaded = True

    def add_processor(self, processor_name, processor_cpath, processor_class):
        assert type(processor_class) is type
        assert processor_name not in self.__name_to_processor
        processor = processor_class(self.__site, processor_cpath)
        self.__name_to_processor[processor_name] = processor
        return processor

    def get_processor(self, processor_name, default=None, error_out=True):
        processor = self.__name_to_processor.get(processor_name, None)
        if processor is None and error_out is True:
            raise SynamicPreProcessorNotFound(f'Processor {processor_name} could not be found')
        elif processor is None:
            return default
        else:
            return processor

    @property
    def pre_processors(self):
        return tuple(self.__name_to_processor.values())

    def __getattr__(self, key):
        return self.get_processor(key, error_out=True)



class _Header:
    def __init__(self, level, text, html_id):
        self.__level = int(level)
        self.__text = text
        self.__html_id = html_id
        self.__children = []

    @property
    def level(self):
        return self.__level

    @property
    def text(self):
        return self.__text

    @property
    def html_id(self):
        return self.__html_id

    @property
    def html(self):
        if not self.children:
            res = "<li class='synamic-toc-header-%s-li'><a href='#%s'>%s</a></li>" % (
                self.level, self.html_id, self.text
            )
        else:
            child_strs = []
            for child in self.__children:
                if len(child.children) == 0:
                    child_strs.append(
                        "<li class='synamic-toc-header-%s'><a href='#%s'>%s</a></li>" % (
                            child.level, child.html_id, child.text)
                    )
                else:
                    child_strs.append(
                        "<li class='synamic-toc-header-%s'><a href='#%s'>%s</a><ul>%s</ul></li>" % (
                            child.level, child.html_id, child.text, child.html)
                    )
            child_str = "\n".join(child_strs)
            res = "<li class='synamic-toc-header-%s-li'><a href='#%s'>%s</a><ul>%s</ul></li>" % (
                self.level, self.html_id, self.text, child_str
            )

        return res

    def add(self, header):
        def findnkill(children, header):
            children_reverse = children.copy(); children_reverse.reverse()
            for child in children_reverse:
                if header.level > child.level:
                    child.children.append(header)
                    return
                else:
                    findnkill(child.children, header)
        if self.level + 1 == header.level:
            self.children.append(header)
        else:
            findnkill(self.children, header)

    @property
    def children(self):
        return self.__children

    def __str__(self):
        return self.html

    def __repr__(self):
        return repr(str(self))


class Toc:
    def __init__(self):
        self.__headers = []

    def __bool__(self):
        return bool(self.__headers)

    def add(self, level, text, html_id):
        level = int(level)
        header = _Header(level, text, html_id)
        if len(self.__headers) == 0 or level == 1:
            self.__headers.append(header)
        else:
            last_header = self.__headers[-1]
            if last_header.level > 1 and last_header.level >= header.level:
                self.__headers.append(header)
            else:
                last_header.add(header)

    @property
    def html(self):
        if len(self.__headers) == 0:
            return ""
        return "<ul class='synamic-toc-primary-ul'>%s<ul>" % "\n".join(header.html for header in self.__headers)

    def __str__(self):
        return self.html

    def __repr__(self):
        return str(self)
